#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Simple agent which repeats back the labels sent to it.

By default, replies with a single random label from the list of labels sent to
it, if any. If the ``label_candidates`` field is set, will fill the ``text_candidates``
field with up to a hundred randomly selected candidates (the first text
candidate is the selected label).

Options:

    ``returnOneRandomAnswer`` -- default ``True``, set to ``False`` to instead
    reply with all labels joined by commas.

    ``cantAnswerPercent`` -- default ``0``, set value in range[0,1] to set
    chance of replying with "I don't know."
"""

import random

from parlai.core.agents import Agent
from parlai.core.message import Message


class RepeatLabelAgent(Agent):
    @staticmethod
    def add_cmdline_args(argparser):
        group = argparser.add_argument_group('RepeatLabel Arguments')
        group.add_argument(
            '--return_one_random_answer',
            type='bool',
            default=True,
            help='return one answer from the set of labels',
        )
        group.add_argument(
            '--cant_answer_percent',
            type=float,
            default=0,
            help='set value in range[0,1] to set chance of '
            'replying with special message',
        )
        group.add_argument(
            '--cant_answer_message',
            type=str,
            default="I don't know.",
            help='Message sent when the model cannot answer',
        )

    def __init__(self, opt, shared=None):
        super().__init__(opt)
        self.returnOneRandomAnswer = opt.get('return_one_random_answer', True)
        self.cantAnswerPercent = opt.get('cant_answer_percent', 0)
        self.cantAnswerMessage = opt.get('cant_answer_message', "I don't know.")
        self.id = 'RepeatLabelAgent'

    def act(self):
        obs = self.observation
        if obs is None:
            return {'text': 'Nothing to repeat yet.'}
        reply = {}
        reply['id'] = self.getID()
        labels = obs.get('labels', obs.get('eval_labels', None))
        if labels:
            if random.random() >= self.cantAnswerPercent:
                if self.returnOneRandomAnswer:
                    reply['text'] = labels[random.randrange(len(labels))]
                else:
                    reply['text'] = ', '.join(labels)
            else:
                # Some 'self.cantAnswerPercent' percentage of the time
                # the agent does not answer.
                reply['text'] = self.cantAnswerMessage
        else:
            reply['text'] = self.cantAnswerMessage
        reply['episode_done'] = False
        return Message(reply)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.teachers import FixedDialogTeacher
from parlai.core.message import Message
from .build import build
import os
import json


def _path(opt):
    # build the data if it does not exist
    build(opt)

    # set up path to data (specific to each dataset)
    jsons_path = os.path.join(opt['datapath'], 'airdialogue_data', 'airdialogue')
    return jsons_path


class AirDialogueTeacher(FixedDialogTeacher):
    """
    AirDialogue Teacher.

    This also contains other files (dev_kb.json, train_kb.json) with flight data about
    return flights, price, connections, flight airlines, departure times, and other
    flight information. More information and related paper can be found at
    <https://github.com/google/airdialogue>.
    """

    def __init__(self, opt, shared=None):
        super().__init__(opt, shared)
        jsons_path = _path(opt)
        self.datatype = opt['datatype'].split(':')[0]
        if shared is not None:
            self.messages = shared['messages']
        else:
            self.messages = []
            self._setup_data(jsons_path)
        # self._setup_data(jsons_path)
        self.id = 'airdialogue'
        self.reset()

    def _setup_data(self, jsons_path):
        train_path = os.path.join(jsons_path, 'train_data.json')
        test_valid_path = os.path.join(jsons_path, 'dev_data.json')
        if self.datatype.startswith('test'):
            self.save_messages_from_path(test_valid_path)
            self.messages = self.messages[len(self.messages) // 2 :]
        elif self.datatype.startswith('valid'):
            self.save_messages_from_path(test_valid_path)
            self.messages = self.messages[: len(self.messages) // 2]
        else:
            self.save_messages_from_path(train_path)

    def save_messages_from_path(self, json_path):
        with open(json_path) as f:
            for line in f:
                if len(line) > 1:
                    self.messages.append(json.loads(line)['dialogue'])

    def share(self):
        shared = super().share()
        shared['messages'] = self.messages
        return shared

    def num_examples(self):
        examples = 0
        for data in self.messages:
            examples += len(data) // 2
        return examples

    def num_episodes(self):
        return len(self.messages)

    def get(self, episode_idx, entry_idx=0):
        log_idx = entry_idx * 2
        entry = self.messages[episode_idx][log_idx]
        entry = entry.split(': ')[1]
        last_backnforth_idx = len(self.messages[episode_idx]) - 2
        # sometimes the first speaker is at the end with no reply
        if len(self.messages[episode_idx]) % 2 == 1:
            last_backnforth_idx -= 1
        episode_done = log_idx >= last_backnforth_idx
        label_text = self.messages[episode_idx][log_idx + 1]
        label_text = label_text.split(': ')[1]
        labels = [label_text]
        action = {
            'id': self.id,
            'text': entry,
            'episode_done': episode_done,
            'labels': labels,
        }
        return Message(action)


class DefaultTeacher(AirDialogueTeacher):
    pass



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
BART Module.
"""
import torch
import torch.nn.functional as F
from typing import Tuple, Any

from parlai.agents.transformer.modules import TransformerGeneratorModel


class BartModel(TransformerGeneratorModel):
    """
    BART Model.
    """

    def output(self, tensor: torch.Tensor) -> torch.Tensor:
        """
        Compute output logits.

        Override standard TGM output to _not_ prevent generation of BOS.
        """
        # project back to vocabulary
        output = F.linear(tensor, self.embeddings.weight)
        return output

    def decode_forced(
        self, encoder_states: Tuple[Any, ...], ys: torch.LongTensor
    ) -> Tuple[torch.FloatTensor, torch.LongTensor]:
        """
        Decode with a fixed, true sequence, computing loss.

        Overriding `TGM.decode_forced` to bypass assertion that BOS is not present, and
        additionally insert EOS as first token
        """
        bsz = ys.size(0)
        seqlen = ys.size(1)
        inputs = ys.narrow(1, 0, seqlen - 1)
        inputs = torch.cat(
            [
                torch.LongTensor([self.END_IDX]).detach().expand(bsz, 1).to(inputs),
                inputs,
            ],
            1,
        )
        latent, _ = self.decoder(inputs, encoder_states)
        logits = self.output(latent)
        _, preds = logits.max(dim=2)
        return logits, preds



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Teachers that wrap around other teachers, for instance, to modify message fields while
keeping the examples/episodes the same.

This is useful when working with agents that expect examples to be in a certain format,
for instance a classifier that classifies the "text" field of a message. The meta-
teachers in this module can be used to avoid writing several different nearly identical
variants of different teachers: for instance, if you want to flatten examples and strip
away all but the previous utterance in the 'text' field for several different teachers,
it would be much easier to do so with one teacher in this module than with a brand new
teacher for each of the original teachers.
"""

import copy

from abc import ABC, abstractmethod
from parlai.core.agents import create_agent_from_shared
from parlai.core.opt import Opt
from parlai.core.teachers import create_task_agent_from_taskname, Teacher


class AbstractWrapperTeacher(Teacher, ABC):
    """
    Abstract teacher that will wrap around another teacher and allow for manipulating
    the fields returned by the inner teacher.
    """

    @classmethod
    def add_cmdline_args(cls, parser):
        agent = parser.add_argument_group('AbstractWrapper args')
        agent.add_argument(
            '-wt',
            '--wrapper-task',
            type=str,
            help='The task whose fields will be manipulated.',
        )
        known_args, _ = parser.parse_known_args(nohelp=True)
        parser.add_task_args(known_args.wrapper_task)

    def __init__(self, opt: Opt, shared=None):
        if ',' in opt['task']:
            raise ValueError(
                'AbstractWrapperTeacher cannot be used with multiple tasks!'
            )
        self.id = opt['task']
        self.opt = opt
        if shared:
            self.task = create_agent_from_shared(shared['task'])
        else:
            opt_singletask = copy.deepcopy(opt)
            opt_singletask['task'] = opt['wrapper_task']
            self.task = create_task_agent_from_taskname(opt_singletask)[0]

    @abstractmethod
    def act(self):
        """
        Act on the previous observation.
        """
        raise NotImplementedError('Abstract class: user must implement act() method')

    def num_examples(self):
        """
        Return the number of examples.
        """
        return self.task.num_examples()

    def num_episodes(self):
        """
        Return the number of episodes.

        Because the dataset is flattened, there will be one episode per example.
        """
        return self.task.num_examples()

    def observe(self, observation):
        """
        Make an observation.
        """
        return self.task.observe(observation)

    def epoch_done(self):
        """
        Return whether the subtask is completed.
        """
        return self.task.epoch_done()

    def report(self):
        """
        Report metrics for the subtask.
        """
        return self.task.report()

    def reset(self):
        """
        Reset the subtask.
        """
        self.task.reset()

    def reset_metrics(self):
        """
        Reset metrics for the subtask.
        """
        self.task.reset_metrics()

    def save(self):
        """
        Save the subtask.
        """
        self.task.save()

    def share(self):
        """
        Share the subtask.
        """
        shared = {}
        shared['class'] = type(self)
        shared['opt'] = self.opt
        shared['task'] = self.task.share()
        return shared


class LabelToTextTeacher(AbstractWrapperTeacher):
    """
    Teacher that will shift message['labels'][0] into message['text'] for whatever task
    is specified with --wrapper-task.

    Because the dialogue history is effectively overwritten by this action, all episodes
    will be flattened into one example each.
    """

    def __init__(self, opt: Opt, shared=None):
        super().__init__(opt, shared)

    def act(self):
        """
        Act on the previous observation.
        """
        act = self.task.act()
        new_act = copy.deepcopy(act)
        if 'labels' in act or 'eval_labels' in act:
            labels_type = 'labels' if 'labels' in act else 'eval_labels'
            labels = act[labels_type]
            if len(labels) != 1:
                raise ValueError('LabelToTextTeacher can only be used with one label!')
            new_act.force_set('text', labels[0])
            new_act.force_set(labels_type, [''])
        else:
            assert 'text' not in act and act['episode_done'] is True
        new_act.force_set('episode_done', True)  # Clear the dialogue history
        return new_act



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.teachers import DialogTeacher
from .build import build

import os
import copy
import csv
import glob


def _path(opt):
    build(opt)

    dt = opt['datatype'].split(':')[0]

    if not (dt == 'train' or dt == 'valid' or dt == 'test'):
        raise RuntimeError('Not valid datatype.')

    suffix = dt

    data_path = os.path.join(opt['datapath'], 'NarrativeQA', 'narrative_qa', suffix)

    return data_path


class SummariesTeacher(DialogTeacher):
    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        data_path = _path(opt)
        opt['datafile'] = data_path
        self.id = 'NarrativeQA'

        super().__init__(opt, shared)

    def setup_data(self, path):
        print('loading data from: ' + path)

        qa_path = os.path.join(path, 'qaps.csv')
        summaries_path = os.path.join(path, 'summaries.csv')

        qa_pairs = dict()

        with open(qa_path, 'r') as f:
            reader = csv.DictReader(f)
            for row in reader:
                if row['document_id'] not in qa_pairs:
                    qa_pairs[row['document_id']] = []
                qa_pairs[row['document_id']].append(row)

        with open(summaries_path, 'r') as f:
            reader = csv.DictReader(f)

            for row in reader:
                info = 'Summary:  %s' % row['summary_tokenized']

                for i, qa in enumerate(qa_pairs[row['document_id']]):
                    question = qa['question_tokenized']
                    answer1 = qa['answer1_tokenized']
                    answer2 = qa['answer2_tokenized']

                    if i == 0:
                        # Prepend start info in first question
                        yield (info + '\n' + question, [answer1, answer2]), True
                    else:
                        yield (question, [answer1, answer2]), False


class DefaultTeacher(DialogTeacher):
    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        data_path = _path(opt)
        opt['datafile'] = data_path
        self.id = 'NarrativeQA'

        super().__init__(opt, shared)

    def setup_data(self, path):
        print('loading data from: ' + path)

        qa_path = os.path.join(path, 'qaps.csv')
        documents_path = os.path.join(path, 'documents.csv')

        stories_base_path = os.path.join(path, '..', 'stories')
        qa_pairs = dict()

        print(
            "%s stories found."
            % len(glob.glob(os.path.join(stories_base_path, "*.content")))
        )

        with open(qa_path, 'r') as f:
            reader = csv.DictReader(f)
            for row in reader:
                if row['document_id'] not in qa_pairs:
                    qa_pairs[row['document_id']] = []
                qa_pairs[row['document_id']].append(row)

        with open(documents_path, 'r') as f:
            reader = csv.DictReader(f)

            for row in reader:
                story_path = os.path.join(
                    stories_base_path, row['document_id'] + '.content'
                )

                if not os.path.exists(story_path):
                    continue

                story = None
                with open(story_path, 'r', encoding='utf-8', errors='ignore') as f:
                    story = f.read().strip()

                info = 'Title:  %s' % row['wiki_title']
                info += '\nKind: %s' % row['kind']
                info += '\nStory url: %s' % row['story_url']
                info += '\nStory start: %s' % row['story_start']
                info += '\nStory end: %s' % row['story_end']
                info += '\nStory: %s' % story

                for i, qa in enumerate(qa_pairs[row['document_id']]):
                    question = qa['question_tokenized']
                    answer1 = qa['answer1_tokenized']
                    answer2 = qa['answer2_tokenized']

                    if i == 0:
                        # Prepend start info in first question
                        yield (info + '\n' + question, [answer1, answer2]), True
                    else:
                        yield (question, [answer1, answer2]), False



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
# Download and build the data if it does not exist.

import parlai.core.build_data as build_data
import os
from parlai.core.build_data import DownloadableFile

RESOURCES = [
    DownloadableFile(
        'https://www.dropbox.com/s/4i9u4y24pt3paba/personalized-dialog-dataset.tar.gz?dl=1',
        'personalized-dialog-dataset.tar.gz',
        '0da3d5ba631d672e9e2d108dfd6721c8201cc41b837425540faba6815c375c52',
    )
]


def build(opt):
    dpath = os.path.join(opt['datapath'], 'personalized-dialog')
    version = None

    if not build_data.built(dpath, version_string=version):
        print('[building data: ' + dpath + ']')
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        # Mark the data as built.
        build_data.mark_done(dpath, version_string=version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict, Any

task_config: Dict[str, Any] = {}


task_config['frontend_version'] = 1

"""A short and descriptive title about the kind of task the HIT contains.
On the Amazon Mechanical Turk web site, the HIT title appears in search results,
and everywhere the HIT is mentioned.
"""
task_config['hit_title'] = 'Ask and answer a question about a paragraph'


"""A description includes detailed information about the kind of task the HIT contains.
On the Amazon Mechanical Turk web site, the HIT description appears in the expanded
view of search results, and in the HIT and assignment screens.
"""
task_config['hit_description'] = 'Ask and answer a question about a paragraph.'


"""One or more words or phrases that describe the HIT, separated by commas.
On MTurk website, these words are used in searches to find HITs.
"""
task_config['hit_keywords'] = 'chat,question,answer'


"""A detailed task description that will be shown on the HIT task preview page
and on the left side of the chat page. Supports HTML formatting.
"""
task_config[
    'task_description'
] = '''\'\'\'
In this task, you will need to ask a question about a paragraph, and then provide your own answer to it.<br><br>
Example:<br><br>
------------------- Task Begin ------------------- <br><br>
<b>QA Collector</b>:<br>
New Haven\'s greatest culinary claim to fame may be its pizza, which has been claimed to be among the best in the country, or even in the world. New Haven-style pizza, called "apizza" (pronounced ah-BEETS, [a'pitts] in the original Italian dialect), made its debut at the iconic Frank Pepe Pizzeria Napoletana (known as Pepe\'s) in 1925. Apizza is baked in coal- or wood-fired brick ovens, and is notable for its thin crust. Apizza may be red (with a tomato-based sauce) or white (with a sauce of garlic and olive oil), and pies ordered "plain" are made without the otherwise customary mozzarella cheese (originally smoked mozzarella, known as "scamorza" in Italian). A white clam pie is a well-known specialty of the restaurants on Wooster Street in the Little Italy section of New Haven, including Pepe\'s and Sally\'s Apizza (which opened in 1938). Modern Apizza on State Street, which opened in 1934, is also well-known.<br><br>Please provide a question given this context.<br><br>
<b>Worker</b>:<br>
What is apizza baked in?<br><br>
<b>QA Collector</b>:<br>
Thanks. And what is the answer to your question?<br><br>
<b>Worker</b>:<br>
It's baked in coal- or wood-fired brick ovens.<br><br>
------------------- Task Done ------------------- <br><br>
If you are ready, please click "Accept HIT" to start this task.
\'\'\''''



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import parlai.mturk.core.mturk_utils as mturk_utils


def main():
    """
    This script should be used to compensate workers that have not recieved proper
    payment for the completion of tasks due to issues on our end.

    It's important to make sure you keep a requester reputation up.
    """
    parser = argparse.ArgumentParser(description='Bonus workers directly')
    parser.add_argument(
        '--sandbox',
        dest='sandbox',
        default=False,
        action='store_true',
        help='Test bonus in sandbox',
    )
    parser.add_argument(
        '--hit-id',
        dest='use_hit_id',
        default=False,
        action='store_true',
        help='Use HIT id instead of assignment id',
    )
    opt = parser.parse_args()
    sandbox = opt.sandbox

    if sandbox:
        print(
            'About to connect to the SANDBOX server. These bonuses will not '
            'actually be paid out.'
        )
    else:
        print(
            'About to connect to the LIVE server. These bonuses will be '
            'deducted from your account balance.'
        )

    mturk_utils.setup_aws_credentials()
    client = mturk_utils.get_mturk_client(sandbox)

    while True:
        worker_id = input("Enter worker id: ")
        if len(worker_id) == 0:
            break
        bonus_amount = input("Enter bonus amount: ")
        if opt.use_hit_id:
            hit_id = input("Enter HIT id: ")
            listed = client.list_assignments_for_hit(HITId=hit_id)
            assignment_id = listed['Assignments'][0]['AssignmentId']
        else:
            assignment_id = input("Enter assignment id: ")
        reason = input("Enter reason: ")
        input(
            "Press enter to bonus {} to worker {} for reason '{}' on "
            "assignment {}.".format(bonus_amount, worker_id, reason, assignment_id)
        )
        resp = client.send_bonus(
            WorkerId=worker_id,
            BonusAmount=str(bonus_amount),
            AssignmentId=assignment_id,
            Reason=reason,
        )
        print(resp)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from parlai.core.params import ParlaiParser
from parlai.mturk.tasks.qualification_flow_example.worlds import (
    QualificationFlowOnboardWorld,
    QualificationFlowSoloWorld,
)
from parlai.mturk.core.mturk_manager import MTurkManager
import parlai.mturk.core.mturk_utils as mturk_utils
from parlai.mturk.tasks.qualification_flow_example.task_config import task_config
import os
import random


def main():
    completed_workers = []
    argparser = ParlaiParser(False, False)
    argparser.add_parlai_data_path()
    argparser.add_mturk_args()
    opt = argparser.parse_args()
    opt['task'] = os.path.basename(os.path.dirname(os.path.abspath(__file__)))
    opt.update(task_config)

    mturk_agent_id = 'Worker'
    mturk_manager = MTurkManager(opt=opt, mturk_agent_ids=[mturk_agent_id])
    mturk_manager.setup_server()
    qual_name = 'ParlAIExcludeQual{}t{}'.format(
        random.randint(10000, 99999), random.randint(10000, 99999)
    )
    qual_desc = (
        'Qualification for a worker not correctly completing the '
        'first iteration of a task. Used to filter to different task pools.'
    )
    qualification_id = mturk_utils.find_or_create_qualification(
        qual_name, qual_desc, opt['is_sandbox']
    )
    print('Created qualification: ', qualification_id)

    def run_onboard(worker):
        world = QualificationFlowOnboardWorld(opt, worker)
        while not world.episode_done():
            world.parley()
        world.shutdown()

    mturk_manager.set_onboard_function(onboard_function=run_onboard)

    try:
        mturk_manager.start_new_run()
        agent_qualifications = [
            {
                'QualificationTypeId': qualification_id,
                'Comparator': 'DoesNotExist',
                'RequiredToPreview': True,
            }
        ]
        mturk_manager.create_hits(qualifications=agent_qualifications)

        mturk_manager.ready_to_accept_workers()

        def check_worker_eligibility(worker):
            return True

        def assign_worker_roles(worker):
            worker[0].id = mturk_agent_id

        global run_conversation

        def run_conversation(mturk_manager, opt, workers):
            mturk_agent = workers[0]
            world = QualificationFlowSoloWorld(
                opt=opt,
                mturk_agent=mturk_agent,
                qualification_id=qualification_id,
                firstTime=(mturk_agent.worker_id not in completed_workers),
            )
            while not world.episode_done():
                world.parley()
            completed_workers.append(mturk_agent.worker_id)
            world.shutdown()
            world.review_work()

        mturk_manager.start_task(
            eligibility_function=check_worker_eligibility,
            assign_role_function=assign_worker_roles,
            task_function=run_conversation,
        )
    except BaseException:
        raise
    finally:
        mturk_utils.delete_qualification(qualification_id, opt['is_sandbox'])
        mturk_manager.expire_all_unassigned_hits()
        mturk_manager.shutdown()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

task_config = {}

"""A short and descriptive title about the kind of task the HIT contains.
On the Amazon Mechanical Turk web site, the HIT title appears in search results,
and everywhere the HIT is mentioned.
"""
task_config['hit_title'] = 'Play a character and chat!'


"""A description includes detailed information about the kind of task the HIT contains.
On the Amazon Mechanical Turk web site, the HIT description appears in the expanded
view of search results, and in the HIT and assignment screens.
"""
task_config[
    'hit_description'
] = 'You will chat to another person while adopting a specific persona.'


"""One or more words or phrases that describe the HIT, separated by commas.
On MTurk website, these words are used in searches to find HITs.
"""
task_config['hit_keywords'] = 'chat,dialog'


"""A detailed task description that will be shown on the HIT task preview page
and on the left side of the chat page. Supports HTML formatting.
"""
task_config[
    'task_description'
] = '''
<br>
<b><h4>Task Description</h4></b>
<br>
(You can keep accepting new HITs after you finish your current one, so keep working on it if you like the task!)
<br>
<b>In this task you will chitchat with another worker, playing the part of a given character.</b>
For example, your given character could be: <br><br> I am a vegetarian. I like swimming. My father used to work for Ford. My favorite band is Maroon5. I got a new job last month, which is about advertising design.
<br>
<br>
Chat with the other person naturally and <b><span style="color:blue">try to get to know each other, i.e.
both ask questions and answer questions of your chat partner
at the same time sticking to your own characters<span style="color:blue"></b>.
<br>
<br>
<b><span style="color:blue">You will get bonus for high quality dialogs.</span></b>
<b>Send short messages, <span style="color:red">max 15 words</span>.</b>
<b>Do not trivially copy the character descriptions into the message.</b>
After a given number of turns, click “DONE" to finish the chat.
There is a <b>2 min</b> time limit for each turn.
<br>
<br>
- Do not reference the task or MTurk itself during the conversation.
<br>
<b><span style="color:red">- No racism, sexism or otherwise offensive comments, or the submission will be rejected and we will report to Amazon.</b></span>
'''



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Context repetition unlikelihood on Wizard of Wikipedia: please see.

<parl.ai/projects/dialogue_unlikelihood>.
"""

from .build import build

VERSION = 'v1.0'


def download(datapath):
    build(datapath, 'rep_wiki_ctxt_v1.tgz', model_type='rep_wiki_ctxt', version=VERSION)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
import os
import time
import json
import threading
import pickle
from unittest import mock
from parlai.mturk.core.worker_manager import WorkerManager
from parlai.mturk.core.agents import MTurkAgent
from parlai.mturk.core.shared_utils import AssignState
from parlai.mturk.core.mturk_manager import MTurkManager
from parlai.mturk.core.socket_manager import SocketManager, Packet
from parlai.core.params import ParlaiParser
from websocket_server import WebsocketServer

import parlai.mturk.core.mturk_manager as MTurkManagerFile
import parlai.mturk.core.data_model as data_model
import parlai.utils.testing as testing_utils

parent_dir = os.path.dirname(os.path.abspath(__file__))
MTurkManagerFile.parent_dir = os.path.dirname(os.path.abspath(__file__))

MTurkManagerFile.mturk_utils = mock.MagicMock()

# Lets ignore the logging part
MTurkManagerFile.shared_utils.print_and_log = mock.MagicMock()

TEST_WORKER_ID_1 = 'TEST_WORKER_ID_1'
TEST_WORKER_ID_2 = 'TEST_WORKER_ID_2'
TEST_WORKER_ID_3 = 'TEST_WORKER_ID_3'
TEST_ASSIGNMENT_ID_1 = 'TEST_ASSIGNMENT_ID_1'
TEST_ASSIGNMENT_ID_2 = 'TEST_ASSIGNMENT_ID_2'
TEST_ASSIGNMENT_ID_3 = 'TEST_ASSIGNMENT_ID_3'
TEST_HIT_ID_1 = 'TEST_HIT_ID_1'
TEST_HIT_ID_2 = 'TEST_HIT_ID_2'
TEST_HIT_ID_3 = 'TEST_HIT_ID_3'
FAKE_ID = 'BOGUS'


def assert_equal_by(val_func, val, max_time):
    start_time = time.time()
    while val_func() != val:
        assert (
            time.time() - start_time < max_time
        ), "Value was not attained in specified time"
        time.sleep(0.1)


class MockSocket:
    def __init__(self):
        self.last_messages = {}
        self.connected = False
        self.disconnected = False
        self.closed = False
        self.ws = None
        self.should_heartbeat = True
        self.fake_workers = []
        self.port = None
        self.launch_socket()
        self.handlers = {}
        while self.ws is None:
            time.sleep(0.05)
        time.sleep(1)

    def send(self, packet):
        self.ws.send_message_to_all(packet)

    def close(self):
        if not self.closed:
            self.ws.server_close()
            self.ws.shutdown()
            self.closed = True

    def do_nothing(self, *args):
        pass

    def launch_socket(self):
        def on_message(client, server, message):
            if self.closed:
                raise Exception('Socket is already closed...')
            if message == '':
                return
            packet_dict = json.loads(message)
            if packet_dict['content']['id'] == 'WORLD_ALIVE':
                self.ws.send_message(client, json.dumps({'type': 'conn_success'}))
                self.connected = True
            elif packet_dict['content']['type'] == 'heartbeat':
                pong = packet_dict['content'].copy()
                pong['type'] = 'pong'
                self.ws.send_message(
                    client,
                    json.dumps(
                        {'type': data_model.SOCKET_ROUTE_PACKET_STRING, 'content': pong}
                    ),
                )
            if 'receiver_id' in packet_dict['content']:
                receiver_id = packet_dict['content']['receiver_id']
                use_func = self.handlers.get(receiver_id, self.do_nothing)
                use_func(packet_dict['content'])

        def on_connect(client, server):
            pass

        def on_disconnect(client, server):
            self.disconnected = True

        def run_socket(*args):
            port = 3030
            while self.port is None:
                try:
                    self.ws = WebsocketServer(port, host='127.0.0.1')
                    self.port = port
                except OSError:
                    port += 1
            self.ws.set_fn_client_left(on_disconnect)
            self.ws.set_fn_new_client(on_connect)
            self.ws.set_fn_message_received(on_message)
            self.ws.run_forever()

        self.listen_thread = threading.Thread(
            target=run_socket, name='Fake-Socket-Thread'
        )
        self.listen_thread.daemon = True
        self.listen_thread.start()


class InitTestMTurkManager(unittest.TestCase):
    """
    Unit tests for MTurkManager setup.
    """

    def setUp(self):
        argparser = ParlaiParser(False, False)
        argparser.add_parlai_data_path()
        argparser.add_mturk_args()
        self.opt = argparser.parse_args([])
        self.opt['task'] = 'unittest'
        self.opt['assignment_duration_in_seconds'] = 6
        self.mturk_agent_ids = ['mturk_agent_1', 'mturk_agent_2']
        self.mturk_manager = MTurkManager(
            opt=self.opt, mturk_agent_ids=self.mturk_agent_ids, is_test=True
        )

    def tearDown(self):
        self.mturk_manager.shutdown()

    def test_init(self):
        manager = self.mturk_manager
        opt = self.opt
        self.assertIsNone(manager.server_url)
        self.assertIsNone(manager.topic_arn)
        self.assertIsNone(manager.server_task_name)
        self.assertIsNone(manager.task_group_id)
        self.assertIsNone(manager.run_id)
        self.assertIsNone(manager.task_files_to_copy)
        self.assertIsNone(manager.onboard_function)
        self.assertIsNone(manager.socket_manager)
        self.assertFalse(manager.is_shutdown)
        self.assertFalse(manager.is_unique)
        self.assertEqual(manager.opt, opt)
        self.assertEqual(manager.mturk_agent_ids, self.mturk_agent_ids)
        self.assertEqual(manager.is_sandbox, opt['is_sandbox'])
        self.assertEqual(manager.num_conversations, opt['num_conversations'])
        self.assertEqual(manager.is_sandbox, opt['is_sandbox'])

        self.assertGreaterEqual(
            manager.required_hits, manager.num_conversations * len(self.mturk_agent_ids)
        )

        self.assertIsNotNone(manager.agent_pool_change_condition)

        self.assertEqual(manager.minimum_messages, opt.get('min_messages', 0))
        self.assertEqual(
            manager.auto_approve_delay, opt.get('auto_approve_delay', 5 * 24 * 3600)
        )
        self.assertEqual(manager.has_time_limit, opt.get('max_time', 0) > 0)
        self.assertIsInstance(manager.worker_manager, WorkerManager)
        self.assertEqual(manager.task_state, manager.STATE_CREATED)

    def test_init_state(self):
        manager = self.mturk_manager
        manager._init_state()
        self.assertEqual(manager.agent_pool, [])
        self.assertEqual(manager.hit_id_list, [])
        self.assertEqual(manager.conversation_index, 0)
        self.assertEqual(manager.started_conversations, 0)
        self.assertEqual(manager.completed_conversations, 0)
        self.assertEqual(manager.task_threads, [])
        self.assertTrue(manager.accepting_workers, True)
        self.assertIsNone(manager.qualifications)
        self.assertGreater(manager.time_limit_checked, time.time() - 1)
        self.assertEqual(manager.task_state, manager.STATE_INIT_RUN)


class TestMTurkManagerUnitFunctions(unittest.TestCase):
    """
    Tests some of the simpler MTurkManager functions that don't require much additional
    state to run.
    """

    def setUp(self):
        self.fake_socket = MockSocket()
        time.sleep(0.1)
        argparser = ParlaiParser(False, False)
        argparser.add_parlai_data_path()
        argparser.add_mturk_args()
        self.opt = argparser.parse_args([])
        self.opt['task'] = 'unittest'
        self.opt['assignment_duration_in_seconds'] = 6
        self.mturk_agent_ids = ['mturk_agent_1', 'mturk_agent_2']
        self.mturk_manager = MTurkManager(
            opt=self.opt, mturk_agent_ids=self.mturk_agent_ids, is_test=True
        )
        self.mturk_manager._init_state()
        self.mturk_manager.port = self.fake_socket.port
        self.agent_1 = MTurkAgent(
            self.opt,
            self.mturk_manager,
            TEST_HIT_ID_1,
            TEST_ASSIGNMENT_ID_1,
            TEST_WORKER_ID_1,
        )
        self.agent_2 = MTurkAgent(
            self.opt,
            self.mturk_manager,
            TEST_HIT_ID_2,
            TEST_ASSIGNMENT_ID_2,
            TEST_WORKER_ID_2,
        )
        self.agent_3 = MTurkAgent(
            self.opt,
            self.mturk_manager,
            TEST_HIT_ID_3,
            TEST_ASSIGNMENT_ID_3,
            TEST_WORKER_ID_3,
        )

    def tearDown(self):
        self.mturk_manager.shutdown()
        self.fake_socket.close()

    def test_move_to_waiting(self):
        manager = self.mturk_manager
        manager.worker_manager.change_agent_conversation = mock.MagicMock()
        manager.socket_manager = mock.MagicMock()
        manager.socket_manager.close_channel = mock.MagicMock()
        manager.force_expire_hit = mock.MagicMock()
        self.agent_1.set_status(AssignState.STATUS_DISCONNECT)
        self.agent_1.reduce_state = mock.MagicMock()
        self.agent_2.reduce_state = mock.MagicMock()
        self.agent_3.reduce_state = mock.MagicMock()

        # Test with a disconnected agent, assert the channel is closed
        manager._move_agents_to_waiting([self.agent_1])
        self.agent_1.reduce_state.assert_called_once()
        manager.socket_manager.close_channel.assert_called_once_with(
            self.agent_1.get_connection_id()
        )
        manager.worker_manager.change_agent_conversation.assert_not_called()
        manager.force_expire_hit.assert_not_called()
        manager.socket_manager.close_channel.reset_mock()

        # Test with a connected agent, should be moved to waiting
        manager._move_agents_to_waiting([self.agent_2])
        self.agent_2.reduce_state.assert_not_called()
        manager.socket_manager.close_channel.assert_not_called()
        manager.worker_manager.change_agent_conversation.assert_called_once()
        args = manager.worker_manager.change_agent_conversation.call_args[1]
        self.assertEqual(args['agent'], self.agent_2)
        self.assertTrue(manager.is_waiting_world(args['conversation_id']))
        self.assertEqual(args['new_agent_id'], 'waiting')
        manager.force_expire_hit.assert_not_called()
        manager.worker_manager.change_agent_conversation.reset_mock()

        # Test when no longer accepting agents
        manager.accepting_workers = False
        manager._move_agents_to_waiting([self.agent_3])
        self.agent_3.reduce_state.assert_not_called()
        manager.socket_manager.close_channel.assert_not_called()
        manager.worker_manager.change_agent_conversation.assert_not_called()
        manager.force_expire_hit.assert_called_once_with(
            self.agent_3.worker_id, self.agent_3.assignment_id
        )

    def test_socket_setup(self):
        """
        Basic socket setup should fail when not in correct state, but succeed otherwise.
        """
        self.mturk_manager.task_state = self.mturk_manager.STATE_CREATED
        with self.assertRaises(AssertionError):
            self.mturk_manager._setup_socket()
        self.mturk_manager.task_group_id = 'TEST_GROUP_ID'
        self.mturk_manager.server_url = 'https://127.0.0.1'
        self.mturk_manager.task_state = self.mturk_manager.STATE_INIT_RUN
        self.mturk_manager._setup_socket()
        self.assertIsInstance(self.mturk_manager.socket_manager, SocketManager)

    def test_worker_alive(self):
        # Setup for test
        manager = self.mturk_manager
        manager.task_group_id = 'TEST_GROUP_ID'
        manager.server_url = 'https://127.0.0.1'
        manager.task_state = manager.STATE_ACCEPTING_WORKERS
        manager._setup_socket()
        manager.force_expire_hit = mock.MagicMock()
        manager._onboard_new_agent = mock.MagicMock()
        manager.socket_manager.open_channel = mock.MagicMock(
            wraps=manager.socket_manager.open_channel
        )
        manager.worker_manager.worker_alive = mock.MagicMock(
            wraps=manager.worker_manager.worker_alive
        )
        open_channel = manager.socket_manager.open_channel
        worker_alive = manager.worker_manager.worker_alive

        # Test no assignment
        alive_packet = Packet(
            '',
            '',
            '',
            '',
            '',
            {
                'worker_id': TEST_WORKER_ID_1,
                'hit_id': TEST_HIT_ID_1,
                'assignment_id': None,
                'conversation_id': None,
            },
            '',
        )
        manager._on_alive(alive_packet)
        open_channel.assert_not_called()
        worker_alive.assert_not_called()
        manager._onboard_new_agent.assert_not_called()

        # Test not accepting workers
        alive_packet = Packet(
            '',
            '',
            '',
            '',
            '',
            {
                'worker_id': TEST_WORKER_ID_1,
                'hit_id': TEST_HIT_ID_1,
                'assignment_id': TEST_ASSIGNMENT_ID_1,
                'conversation_id': None,
            },
            '',
        )
        manager.accepting_workers = False
        manager._on_alive(alive_packet)
        open_channel.assert_called_once_with(TEST_WORKER_ID_1, TEST_ASSIGNMENT_ID_1)
        worker_alive.assert_called_once_with(TEST_WORKER_ID_1)
        worker_state = manager.worker_manager._get_worker(TEST_WORKER_ID_1)
        self.assertIsNotNone(worker_state)
        open_channel.reset_mock()
        worker_alive.reset_mock()
        manager.force_expire_hit.assert_called_once_with(
            TEST_WORKER_ID_1, TEST_ASSIGNMENT_ID_1
        )
        manager._onboard_new_agent.assert_not_called()
        manager.force_expire_hit.reset_mock()

        # Test successful creation
        manager.accepting_workers = True
        manager._on_alive(alive_packet)
        open_channel.assert_called_once_with(TEST_WORKER_ID_1, TEST_ASSIGNMENT_ID_1)
        worker_alive.assert_called_once_with(TEST_WORKER_ID_1)
        manager._onboard_new_agent.assert_called_once()
        manager._onboard_new_agent.reset_mock()
        manager.force_expire_hit.assert_not_called()

        agent = manager.worker_manager.get_agent_for_assignment(TEST_ASSIGNMENT_ID_1)
        self.assertIsInstance(agent, MTurkAgent)
        self.assertEqual(agent.get_status(), AssignState.STATUS_NONE)

        # Reconnect in various conditions
        agent.set_status = mock.MagicMock(wraps=agent.set_status)
        manager._add_agent_to_pool = mock.MagicMock()

        # Reconnect when none state no connection_id
        agent.log_reconnect = mock.MagicMock(wraps=agent.log_reconnect)
        manager._on_alive(alive_packet)
        manager.force_expire_hit.assert_called_once_with(
            TEST_WORKER_ID_1, TEST_ASSIGNMENT_ID_1
        )
        manager.force_expire_hit.reset_mock()
        agent.set_status.assert_not_called()
        manager._add_agent_to_pool.assert_not_called()
        agent.log_reconnect.assert_called_once()
        agent.log_reconnect.reset_mock()

        # Reconnect in None state onboarding conversation_id
        alive_packet = Packet(
            '',
            '',
            '',
            '',
            '',
            {
                'worker_id': TEST_WORKER_ID_1,
                'hit_id': TEST_HIT_ID_1,
                'assignment_id': TEST_ASSIGNMENT_ID_1,
                'conversation_id': 'o_1234',
            },
            '',
        )
        manager._on_alive(alive_packet)
        manager.force_expire_hit.assert_not_called()
        agent.set_status.assert_called_once_with(AssignState.STATUS_ONBOARDING)
        manager._add_agent_to_pool.assert_not_called()
        agent.log_reconnect.assert_called_once()
        agent.log_reconnect.reset_mock()

        # Reconnect in None state waiting conversation_id
        alive_packet = Packet(
            '',
            '',
            '',
            '',
            '',
            {
                'worker_id': TEST_WORKER_ID_1,
                'hit_id': TEST_HIT_ID_1,
                'assignment_id': TEST_ASSIGNMENT_ID_1,
                'conversation_id': 'w_1234',
            },
            '',
        )
        agent.set_status(AssignState.STATUS_NONE)
        agent.set_status.reset_mock()
        manager._on_alive(alive_packet)
        manager.force_expire_hit.assert_not_called()
        agent.set_status.assert_called_once_with(AssignState.STATUS_WAITING)
        manager._add_agent_to_pool.assert_called_once_with(agent)
        manager._add_agent_to_pool.reset_mock()
        agent.log_reconnect.assert_called_once()
        agent.log_reconnect.reset_mock()

        # Reconnect in onboarding with waiting conversation id
        agent.set_status(AssignState.STATUS_ONBOARDING)
        agent.set_status.reset_mock()
        manager._on_alive(alive_packet)
        manager.force_expire_hit.assert_not_called()
        agent.set_status.assert_called_once_with(AssignState.STATUS_WAITING)
        manager._add_agent_to_pool.assert_called_once_with(agent)
        manager._add_agent_to_pool.reset_mock()
        agent.log_reconnect.assert_called_once()
        agent.log_reconnect.reset_mock()

        # Reconnect in onboarding with no conversation id
        alive_packet = Packet(
            '',
            '',
            '',
            '',
            '',
            {
                'worker_id': TEST_WORKER_ID_1,
                'hit_id': TEST_HIT_ID_1,
                'assignment_id': TEST_ASSIGNMENT_ID_1,
                'conversation_id': None,
            },
            '',
        )
        agent.set_status(AssignState.STATUS_ONBOARDING)
        agent.set_status.reset_mock()
        manager._restore_agent_state = mock.MagicMock()
        manager._on_alive(alive_packet)
        manager.force_expire_hit.assert_not_called()
        manager._restore_agent_state.assert_called_once_with(
            TEST_WORKER_ID_1, TEST_ASSIGNMENT_ID_1
        )
        agent.set_status.assert_not_called()
        manager._add_agent_to_pool.assert_not_called()
        agent.log_reconnect.assert_called_once()
        agent.log_reconnect.reset_mock()

        # Reconnect in onboarding but not accepting new workers
        manager.accepting_workers = False
        agent.set_status(AssignState.STATUS_ONBOARDING)
        agent.set_status.reset_mock()
        manager._restore_agent_state = mock.MagicMock()
        manager._on_alive(alive_packet)
        manager.force_expire_hit.assert_called_once_with(
            TEST_WORKER_ID_1, TEST_ASSIGNMENT_ID_1
        )
        manager.force_expire_hit.reset_mock()
        manager._restore_agent_state.assert_not_called()
        agent.set_status.assert_not_called()
        manager._add_agent_to_pool.assert_not_called()
        agent.log_reconnect.assert_called_once()
        agent.log_reconnect.reset_mock()

        # Reconnect in waiting no conv id
        manager.accepting_workers = True
        agent.set_status(AssignState.STATUS_WAITING)
        agent.set_status.reset_mock()
        manager._restore_agent_state = mock.MagicMock()
        manager._on_alive(alive_packet)
        manager.force_expire_hit.assert_not_called()
        manager._restore_agent_state.assert_called_once_with(
            TEST_WORKER_ID_1, TEST_ASSIGNMENT_ID_1
        )
        agent.set_status.assert_not_called()
        manager._add_agent_to_pool.assert_called_once()
        manager._add_agent_to_pool.reset_mock()
        agent.log_reconnect.assert_called_once()
        agent.log_reconnect.reset_mock()

        # Reconnect in waiting with conv id
        alive_packet = Packet(
            '',
            '',
            '',
            '',
            '',
            {
                'worker_id': TEST_WORKER_ID_1,
                'hit_id': TEST_HIT_ID_1,
                'assignment_id': TEST_ASSIGNMENT_ID_1,
                'conversation_id': 'w_1234',
            },
            '',
        )
        agent.set_status(AssignState.STATUS_WAITING)
        agent.set_status.reset_mock()
        manager._restore_agent_state = mock.MagicMock()
        manager._on_alive(alive_packet)
        manager.force_expire_hit.assert_not_called()
        manager._restore_agent_state.assert_not_called()
        agent.set_status.assert_not_called()
        manager._add_agent_to_pool.assert_called_once()
        manager._add_agent_to_pool.reset_mock()
        agent.log_reconnect.assert_called_once()
        agent.log_reconnect.reset_mock()

        # Reconnect in waiting with task id
        alive_packet = Packet(
            '',
            '',
            '',
            '',
            '',
            {
                'worker_id': TEST_WORKER_ID_1,
                'hit_id': TEST_HIT_ID_1,
                'assignment_id': TEST_ASSIGNMENT_ID_1,
                'conversation_id': 't_1234',
            },
            '',
        )
        agent.set_status(AssignState.STATUS_WAITING)
        agent.set_status.reset_mock()
        manager._restore_agent_state = mock.MagicMock()
        manager._on_alive(alive_packet)
        manager.force_expire_hit.assert_not_called()
        manager._restore_agent_state.assert_not_called()
        agent.set_status.assert_called_with(AssignState.STATUS_IN_TASK)
        manager._add_agent_to_pool.assert_not_called()
        agent.log_reconnect.assert_called_once()
        agent.log_reconnect.reset_mock()

        # Test active convos failure
        agent.set_status(AssignState.STATUS_IN_TASK)
        agent.set_status.reset_mock()
        alive_packet = Packet(
            '',
            '',
            '',
            '',
            '',
            {
                'worker_id': TEST_WORKER_ID_1,
                'hit_id': TEST_HIT_ID_1,
                'assignment_id': TEST_ASSIGNMENT_ID_2,
                'conversation_id': None,
            },
            '',
        )
        manager.opt['allowed_conversations'] = 1
        manager._on_alive(alive_packet)
        agent.set_status.assert_not_called()
        manager.force_expire_hit.assert_called_once()
        manager._onboard_new_agent.assert_not_called()
        manager.force_expire_hit.reset_mock()

        # Test uniqueness failed
        agent.set_status(AssignState.STATUS_DONE)
        agent.set_status.reset_mock()
        alive_packet = Packet(
            '',
            '',
            '',
            '',
            '',
            {
                'worker_id': TEST_WORKER_ID_1,
                'hit_id': TEST_HIT_ID_1,
                'assignment_id': TEST_ASSIGNMENT_ID_2,
                'conversation_id': None,
            },
            '',
        )
        manager.is_unique = True
        manager._on_alive(alive_packet)
        agent.set_status.assert_not_called()
        manager.force_expire_hit.assert_called_once()
        manager._onboard_new_agent.assert_not_called()
        manager.force_expire_hit.reset_mock()

        # Test in task reconnects
        agent.set_status(AssignState.STATUS_IN_TASK)
        agent.set_status.reset_mock()
        alive_packet = Packet(
            '',
            '',
            '',
            '',
            '',
            {
                'worker_id': TEST_WORKER_ID_1,
                'hit_id': TEST_HIT_ID_1,
                'assignment_id': TEST_ASSIGNMENT_ID_1,
                'conversation_id': None,
            },
            '',
        )
        manager._on_alive(alive_packet)
        manager._restore_agent_state.assert_called_once_with(
            TEST_WORKER_ID_1, TEST_ASSIGNMENT_ID_1
        )
        agent.set_status.assert_not_called()
        manager._add_agent_to_pool.assert_not_called()
        agent.log_reconnect.assert_called_once()
        agent.log_reconnect.reset_mock()

        # Test all final states
        for use_state in [
            AssignState.STATUS_DISCONNECT,
            AssignState.STATUS_DONE,
            AssignState.STATUS_EXPIRED,
            AssignState.STATUS_RETURNED,
            AssignState.STATUS_PARTNER_DISCONNECT,
        ]:
            manager.send_command = mock.MagicMock()
            agent.set_status(use_state)
            agent.set_status.reset_mock()
            manager._on_alive(alive_packet)
            agent.set_status.assert_not_called()
            manager._add_agent_to_pool.assert_not_called()
            manager.force_expire_hit.assert_not_called()
            manager.send_command.assert_called_once()

    def test_mturk_messages(self):
        """
        Ensure incoming messages work as expected.
        """
        # Setup for test
        manager = self.mturk_manager
        manager.task_group_id = 'TEST_GROUP_ID'
        manager.server_url = 'https://127.0.0.1'
        manager.task_state = manager.STATE_ACCEPTING_WORKERS
        manager._setup_socket()
        manager.force_expire_hit = mock.MagicMock()
        manager._on_socket_dead = mock.MagicMock()

        alive_packet = Packet(
            '',
            '',
            '',
            '',
            '',
            {
                'worker_id': TEST_WORKER_ID_1,
                'hit_id': TEST_HIT_ID_1,
                'assignment_id': TEST_ASSIGNMENT_ID_1,
                'conversation_id': None,
            },
            '',
        )
        manager._on_alive(alive_packet)
        agent = manager.worker_manager.get_agent_for_assignment(TEST_ASSIGNMENT_ID_1)
        self.assertIsInstance(agent, MTurkAgent)
        self.assertEqual(agent.get_status(), AssignState.STATUS_NONE)
        agent.set_hit_is_abandoned = mock.MagicMock()

        # Test SNS_ASSIGN_ABANDONDED
        message_packet = Packet(
            '',
            '',
            '',
            '',
            TEST_ASSIGNMENT_ID_1,
            {'text': MTurkManagerFile.SNS_ASSIGN_ABANDONDED},
            '',
        )
        manager._handle_mturk_message(message_packet)
        agent.set_hit_is_abandoned.assert_called_once()
        agent.set_hit_is_abandoned.reset_mock()
        manager._on_socket_dead.assert_called_once_with(
            TEST_WORKER_ID_1, TEST_ASSIGNMENT_ID_1
        )
        manager._on_socket_dead.reset_mock()

        # Test SNS_ASSIGN_RETURNED
        message_packet = Packet(
            '',
            '',
            '',
            '',
            TEST_ASSIGNMENT_ID_1,
            {'text': MTurkManagerFile.SNS_ASSIGN_RETURNED},
            '',
        )
        agent.hit_is_returned = False
        manager._handle_mturk_message(message_packet)
        manager._on_socket_dead.assert_called_once_with(
            TEST_WORKER_ID_1, TEST_ASSIGNMENT_ID_1
        )
        manager._on_socket_dead.reset_mock()
        self.assertTrue(agent.hit_is_returned)

        # Test SNS_ASSIGN_SUBMITTED
        message_packet = Packet(
            '',
            '',
            '',
            '',
            TEST_ASSIGNMENT_ID_1,
            {'text': MTurkManagerFile.SNS_ASSIGN_SUBMITTED},
            '',
        )
        agent.hit_is_complete = False
        manager._handle_mturk_message(message_packet)
        manager._on_socket_dead.assert_not_called()
        self.assertTrue(agent.hit_is_complete)

    def test_new_message(self):
        """
        test on_new_message.
        """
        alive_packet = Packet(
            '',
            TEST_WORKER_ID_1,
            '',
            '',
            '',
            {
                'worker_id': TEST_WORKER_ID_1,
                'hit_id': TEST_HIT_ID_1,
                'assignment_id': TEST_ASSIGNMENT_ID_1,
                'conversation_id': None,
            },
            '',
        )

        message_packet = Packet(
            '',
            '',
            MTurkManagerFile.AMAZON_SNS_NAME,
            '',
            TEST_ASSIGNMENT_ID_1,
            {'text': MTurkManagerFile.SNS_ASSIGN_SUBMITTED},
            '',
        )

        manager = self.mturk_manager
        manager._handle_mturk_message = mock.MagicMock()
        manager.worker_manager.route_packet = mock.MagicMock()

        # test mturk message
        manager._on_new_message(alive_packet)
        manager._handle_mturk_message.assert_not_called()
        manager.worker_manager.route_packet.assert_called_once_with(alive_packet)
        manager.worker_manager.route_packet.reset_mock()

        # test non-mturk message
        manager._on_new_message(message_packet)
        manager._handle_mturk_message.assert_called_once_with(message_packet)
        manager.worker_manager.route_packet.assert_not_called()

    @testing_utils.retry()
    def test_onboarding_function(self):
        manager = self.mturk_manager
        manager.onboard_function = mock.MagicMock()
        manager.worker_manager.change_agent_conversation = mock.MagicMock()
        manager._move_agents_to_waiting = mock.MagicMock()
        manager.worker_manager.get_agent_for_assignment = mock.MagicMock(
            return_value=self.agent_1
        )

        onboard_threads = manager.assignment_to_onboard_thread
        did_launch = manager._onboard_new_agent(self.agent_1)
        assert_equal_by(onboard_threads[self.agent_1.assignment_id].isAlive, True, 0.2)
        time.sleep(0.1)
        manager.worker_manager.change_agent_conversation.assert_called_once()
        manager.worker_manager.change_agent_conversation.reset_mock()
        manager.onboard_function.assert_not_called()
        self.assertTrue(did_launch)

        # Thread will be waiting for agent_1 status to go to ONBOARDING, ensure
        # won't start new thread on a repeat call when first still alive
        did_launch = manager._onboard_new_agent(self.agent_1)
        time.sleep(0.2)
        manager.worker_manager.change_agent_conversation.assert_not_called()
        manager.worker_manager.get_agent_for_assignment.assert_not_called()
        manager.onboard_function.assert_not_called()
        self.assertFalse(did_launch)

        # Advance the worker to simulate a connection, assert onboarding goes
        self.agent_1.set_status(AssignState.STATUS_ONBOARDING)
        assert_equal_by(onboard_threads[self.agent_1.assignment_id].isAlive, False, 0.6)
        manager.onboard_function.assert_called_with(self.agent_1)
        manager._move_agents_to_waiting.assert_called_once()

        # Try to launch a new onboarding world for the same agent still in
        # onboarding, assert that this call is ignored.
        did_launch = manager._onboard_new_agent(self.agent_1)
        self.assertFalse(did_launch)

        # Try to launch with an agent that was in none but supposedly launched
        # before
        self.agent_1.set_status(AssignState.STATUS_NONE)
        did_launch = manager._onboard_new_agent(self.agent_1)
        self.assertTrue(did_launch)
        self.agent_1.set_status(AssignState.STATUS_ONBOARDING)

    def test_agents_incomplete(self):
        agents = [self.agent_1, self.agent_2, self.agent_3]
        manager = self.mturk_manager
        self.assertFalse(manager._no_agents_incomplete(agents))
        self.agent_1.set_status(AssignState.STATUS_DISCONNECT)
        self.assertFalse(manager._no_agents_incomplete(agents))
        self.agent_2.set_status(AssignState.STATUS_DONE)
        self.assertFalse(manager._no_agents_incomplete(agents))
        self.agent_3.set_status(AssignState.STATUS_PARTNER_DISCONNECT)
        self.assertFalse(manager._no_agents_incomplete(agents))
        self.agent_1.set_status(AssignState.STATUS_DONE)
        self.assertFalse(manager._no_agents_incomplete(agents))
        self.agent_3.set_status(AssignState.STATUS_DONE)
        self.assertTrue(manager._no_agents_incomplete(agents))

    def test_world_types(self):
        onboard_type = 'o_12345'
        waiting_type = 'w_12345'
        task_type = 't_12345'
        garbage_type = 'g_12345'
        manager = self.mturk_manager
        self.assertTrue(manager.is_onboarding_world(onboard_type))
        self.assertTrue(manager.is_task_world(task_type))
        self.assertTrue(manager.is_waiting_world(waiting_type))
        for world_type in [waiting_type, task_type, garbage_type]:
            self.assertFalse(manager.is_onboarding_world(world_type))
        for world_type in [onboard_type, task_type, garbage_type]:
            self.assertFalse(manager.is_waiting_world(world_type))
        for world_type in [waiting_type, onboard_type, garbage_type]:
            self.assertFalse(manager.is_task_world(world_type))

    def test_turk_timeout(self):
        """
        Timeout should send expiration message to worker and be treated as a disconnect
        event.
        """
        manager = self.mturk_manager
        manager.force_expire_hit = mock.MagicMock()
        manager._handle_agent_disconnect = mock.MagicMock()

        manager.handle_turker_timeout(TEST_WORKER_ID_1, TEST_ASSIGNMENT_ID_1)
        manager.force_expire_hit.assert_called_once()
        call_args = manager.force_expire_hit.call_args
        self.assertEqual(call_args[0][0], TEST_WORKER_ID_1)
        self.assertEqual(call_args[0][1], TEST_ASSIGNMENT_ID_1)
        manager._handle_agent_disconnect.assert_called_once_with(
            TEST_WORKER_ID_1, TEST_ASSIGNMENT_ID_1
        )

    @testing_utils.retry()
    def test_wait_for_task_expirations(self):
        """
        Ensure waiting for expiration time actually works out.
        """
        manager = self.mturk_manager
        manager.opt['assignment_duration_in_seconds'] = 0.5
        manager.expire_all_unassigned_hits = mock.MagicMock()
        manager.hit_id_list = [1, 2, 3]

        def run_task_wait():
            manager._wait_for_task_expirations()

        wait_thread = threading.Thread(target=run_task_wait, daemon=True)
        wait_thread.start()
        time.sleep(0.1)
        self.assertTrue(wait_thread.isAlive())
        assert_equal_by(wait_thread.isAlive, False, 0.6)

    def test_mark_workers_done(self):
        manager = self.mturk_manager
        manager.give_worker_qualification = mock.MagicMock()
        manager._log_working_time = mock.MagicMock()
        manager.has_time_limit = False

        # Assert finality doesn't change
        self.agent_1.set_status(AssignState.STATUS_DISCONNECT)
        manager.mark_workers_done([self.agent_1])
        self.assertEqual(AssignState.STATUS_DISCONNECT, self.agent_1.get_status())

        # assert uniqueness works as expected
        manager.is_unique = True
        with self.assertRaises(AssertionError):
            manager.mark_workers_done([self.agent_2])
        manager.give_worker_qualification.assert_not_called()
        manager.unique_qual_name = 'fake_qual_name'
        manager.mark_workers_done([self.agent_2])
        manager.give_worker_qualification.assert_called_once_with(
            self.agent_2.worker_id, 'fake_qual_name'
        )
        self.assertEqual(self.agent_2.get_status(), AssignState.STATUS_DONE)
        manager.is_unique = False

        # Ensure working time is called if it's set
        manager.has_time_limit = True
        manager.mark_workers_done([self.agent_3])
        self.assertEqual(self.agent_3.get_status(), AssignState.STATUS_DONE)
        manager._log_working_time.assert_called_once_with(self.agent_3)


class TestMTurkManagerPoolHandling(unittest.TestCase):
    def setUp(self):
        argparser = ParlaiParser(False, False)
        argparser.add_parlai_data_path()
        argparser.add_mturk_args()
        self.opt = argparser.parse_args([])
        self.opt['task'] = 'unittest'
        self.opt['assignment_duration_in_seconds'] = 6
        self.mturk_agent_ids = ['mturk_agent_1', 'mturk_agent_2']
        self.mturk_manager = MTurkManager(
            opt=self.opt, mturk_agent_ids=self.mturk_agent_ids, is_test=True
        )
        self.mturk_manager._init_state()
        self.agent_1 = MTurkAgent(
            self.opt,
            self.mturk_manager,
            TEST_HIT_ID_1,
            TEST_ASSIGNMENT_ID_1,
            TEST_WORKER_ID_1,
        )
        self.agent_2 = MTurkAgent(
            self.opt,
            self.mturk_manager,
            TEST_HIT_ID_2,
            TEST_ASSIGNMENT_ID_2,
            TEST_WORKER_ID_2,
        )
        self.agent_3 = MTurkAgent(
            self.opt,
            self.mturk_manager,
            TEST_HIT_ID_3,
            TEST_ASSIGNMENT_ID_3,
            TEST_WORKER_ID_3,
        )

    def tearDown(self):
        self.mturk_manager.shutdown()

    def test_pool_add_get_remove_and_expire(self):
        """
        Ensure the pool properly adds and releases workers.
        """
        all_are_eligible = {'multiple': True, 'func': lambda workers: workers}
        manager = self.mturk_manager

        # Test empty pool
        pool = manager._get_unique_pool(all_are_eligible)
        self.assertEqual(pool, [])

        # Test pool add and get
        manager._add_agent_to_pool(self.agent_1)
        manager._add_agent_to_pool(self.agent_2)
        manager._add_agent_to_pool(self.agent_3)
        self.assertListEqual(
            manager._get_unique_pool(all_are_eligible),
            [self.agent_1, self.agent_2, self.agent_3],
        )

        # Test extra add to pool has no effect
        manager._add_agent_to_pool(self.agent_1)
        self.assertListEqual(
            manager._get_unique_pool(all_are_eligible),
            [self.agent_1, self.agent_2, self.agent_3],
        )

        # Test remove from the pool works:
        manager._remove_from_agent_pool(self.agent_2)
        self.assertListEqual(
            manager._get_unique_pool(all_are_eligible), [self.agent_1, self.agent_3]
        )

        # Test repeated remove fails
        with self.assertRaises(AssertionError):
            manager._remove_from_agent_pool(self.agent_2)

        # Test eligibility function
        second_worker_only = {'multiple': True, 'func': lambda workers: [workers[1]]}
        self.assertListEqual(
            manager._get_unique_pool(second_worker_only), [self.agent_3]
        )

        # Test single eligibility function
        only_agent_1 = {
            'multiple': False,
            'func': lambda worker: worker is self.agent_1,
        }
        self.assertListEqual(manager._get_unique_pool(only_agent_1), [self.agent_1])

        # Test expiration of pool
        manager.force_expire_hit = mock.MagicMock()

        manager._expire_agent_pool()
        manager.force_expire_hit.assert_any_call(
            self.agent_1.worker_id, self.agent_1.assignment_id
        )
        manager.force_expire_hit.assert_any_call(
            self.agent_3.worker_id, self.agent_3.assignment_id
        )
        pool = manager._get_unique_pool(all_are_eligible)
        self.assertEqual(pool, [])

        # Test adding two agents from the same worker
        self.agent_2.worker_id = self.agent_1.worker_id
        manager._add_agent_to_pool(self.agent_1)
        manager._add_agent_to_pool(self.agent_2)
        # both workers are in the pool
        self.assertListEqual(manager.agent_pool, [self.agent_1, self.agent_2])
        # Only one worker per unique list though
        manager.is_sandbox = False
        self.assertListEqual(manager._get_unique_pool(all_are_eligible), [self.agent_1])


class TestMTurkManagerTimeHandling(unittest.TestCase):
    def setUp(self):
        argparser = ParlaiParser(False, False)
        argparser.add_parlai_data_path()
        argparser.add_mturk_args()
        self.opt = argparser.parse_args([])
        self.opt['task'] = 'unittest'
        self.opt['assignment_duration_in_seconds'] = 6
        self.mturk_agent_ids = ['mturk_agent_1', 'mturk_agent_2']
        self.mturk_manager = MTurkManager(
            opt=self.opt, mturk_agent_ids=self.mturk_agent_ids, is_test=True
        )
        self.mturk_manager.time_limit_checked = time.time()
        self.mturk_manager.worker_manager.un_time_block_workers = mock.MagicMock()
        self.mturk_manager.worker_manager.time_block_worker = mock.MagicMock()
        self.old_time = MTurkManagerFile.time
        MTurkManagerFile.time = mock.MagicMock()
        MTurkManagerFile.time.time = mock.MagicMock(return_value=0)
        self.agent_1 = MTurkAgent(
            self.opt,
            self.mturk_manager,
            TEST_HIT_ID_1,
            TEST_ASSIGNMENT_ID_1,
            TEST_WORKER_ID_1,
        )
        self.agent_2 = MTurkAgent(
            self.opt,
            self.mturk_manager,
            TEST_HIT_ID_2,
            TEST_ASSIGNMENT_ID_2,
            TEST_WORKER_ID_2,
        )

    def tearDown(self):
        self.mturk_manager.shutdown()
        MTurkManagerFile.time = self.old_time

    def test_create_work_time_file(self):
        manager = self.mturk_manager
        manager._should_use_time_logs = mock.MagicMock(return_value=True)

        file_path = os.path.join(parent_dir, MTurkManagerFile.TIME_LOGS_FILE_NAME)
        file_lock = os.path.join(parent_dir, MTurkManagerFile.TIME_LOGS_FILE_LOCK)
        # No lock should exist already
        self.assertFalse(os.path.exists(file_lock))

        # open the work time file, ensure it was just updated
        MTurkManagerFile.time.time = mock.MagicMock(return_value=42424242)
        manager._reset_time_logs(force=True)
        with open(file_path, 'rb+') as time_log_file:
            existing_times = pickle.load(time_log_file)
            self.assertEqual(existing_times['last_reset'], 42424242)
            self.assertEqual(len(existing_times), 1)

        # Try to induce a check, ensure it doesn't fire because too recent
        MTurkManagerFile.time.time = mock.MagicMock(return_value=(60 * 60 * 24 * 1000))
        manager._check_time_limit()
        manager.worker_manager.un_time_block_workers.assert_not_called()

        # Try to induce a check, ensure it doesn't fire because outside of 30
        # minute window
        MTurkManagerFile.time.time = mock.MagicMock(
            return_value=(60 * 60 * 24 * 1000) + (60 * 40)
        )
        manager.time_limit_checked = 0
        manager._check_time_limit()
        manager.worker_manager.un_time_block_workers.assert_not_called()

        # Induce a check
        MTurkManagerFile.time.time = mock.MagicMock(return_value=(60 * 60 * 24 * 1000))
        manager._check_time_limit()
        self.assertEqual(manager.time_limit_checked, (60 * 60 * 24 * 1000))

    def test_add_to_work_time_file_and_block(self):
        manager = self.mturk_manager
        self.agent_1.creation_time = 1000
        self.agent_2.creation_time = 1000
        manager.opt['max_time'] = 10000
        # Ensure a worker below the time limit isn't blocked
        MTurkManagerFile.time.time = mock.MagicMock(return_value=10000)
        self.mturk_manager._should_use_time_logs = mock.MagicMock(return_value=True)
        manager._log_working_time(self.agent_1)
        manager.worker_manager.time_block_worker.assert_not_called()

        # Ensure a worker above the time limit is blocked
        MTurkManagerFile.time.time = mock.MagicMock(return_value=100000)
        manager._log_working_time(self.agent_2)
        manager.worker_manager.time_block_worker.assert_called_with(
            self.agent_2.worker_id
        )

        # Ensure on a (forced) reset all workers are freed
        manager._reset_time_logs(force=True)
        manager.worker_manager.un_time_block_workers.assert_called_once()
        args = manager.worker_manager.un_time_block_workers.call_args
        worker_list = args[0][0]
        self.assertIn(self.agent_1.worker_id, worker_list)
        self.assertIn(self.agent_2.worker_id, worker_list)


class TestMTurkManagerLifecycleFunctions(unittest.TestCase):
    def setUp(self):
        self.fake_socket = MockSocket()
        argparser = ParlaiParser(False, False)
        argparser.add_parlai_data_path()
        argparser.add_mturk_args()
        self.opt = argparser.parse_args([])
        self.opt['task'] = 'unittest'
        self.opt['task_description'] = 'Test task description'
        self.opt['assignment_duration_in_seconds'] = 6
        self.mturk_agent_ids = ['mturk_agent_1', 'mturk_agent_2']
        self.mturk_manager = MTurkManager(
            opt=self.opt, mturk_agent_ids=self.mturk_agent_ids, is_test=True
        )
        MTurkManagerFile.server_utils.delete_server = mock.MagicMock()

    def tearDown(self):
        self.mturk_manager.shutdown()
        self.fake_socket.close()

    @testing_utils.retry()
    def test_full_lifecycle(self):
        manager = self.mturk_manager
        server_url = 'https://fake_server_url'
        topic_arn = 'aws_topic_arn'
        mturk_page_url = 'https://test_mturk_page_url'
        MTurkManagerFile.server_utils.setup_server = mock.MagicMock(
            return_value=server_url
        )
        MTurkManagerFile.server_utils.setup_legacy_server = mock.MagicMock(
            return_value=server_url
        )

        # Currently in state created. Try steps that are too soon to work
        with self.assertRaises(AssertionError):
            manager.start_new_run()
        with self.assertRaises(AssertionError):
            manager.start_task(None, None, None)

        # Setup the server but fail due to insufficent funds
        manager.opt['local'] = True
        MTurkManagerFile.input = mock.MagicMock()
        MTurkManagerFile.mturk_utils.setup_aws_credentials = mock.MagicMock()
        MTurkManagerFile.mturk_utils.check_mturk_balance = mock.MagicMock(
            return_value=False
        )
        MTurkManagerFile.mturk_utils.calculate_mturk_cost = mock.MagicMock(
            return_value=10
        )
        with self.assertRaises(SystemExit):
            manager.setup_server()

        MTurkManagerFile.mturk_utils.setup_aws_credentials.assert_called_once()
        MTurkManagerFile.mturk_utils.check_mturk_balance.assert_called_once()
        MTurkManagerFile.input.assert_called()
        # Two calls to to input if local is set
        self.assertEqual(len(MTurkManagerFile.input.call_args_list), 2)

        # Test successful setup
        manager.opt['local'] = False
        MTurkManagerFile.input.reset_mock()
        MTurkManagerFile.mturk_utils.check_mturk_balance = mock.MagicMock(
            return_value=True
        )
        MTurkManagerFile.mturk_utils.create_hit_config = mock.MagicMock()
        manager.setup_server()
        # Copy one file for cover page, 2 workers, and 1 onboarding
        self.assertEqual(len(manager.task_files_to_copy), 4)
        self.assertEqual(manager.server_url, server_url)
        self.assertIn('unittest', manager.server_task_name)
        MTurkManagerFile.input.assert_called_once()
        MTurkManagerFile.mturk_utils.check_mturk_balance.assert_called_once()
        MTurkManagerFile.mturk_utils.create_hit_config.assert_called_once()
        self.assertEqual(manager.task_state, manager.STATE_SERVER_ALIVE)

        # Start a new run
        MTurkManagerFile.mturk_utils.setup_sns_topic = mock.MagicMock(
            return_value=topic_arn
        )
        manager._init_state = mock.MagicMock(wraps=manager._init_state)
        manager.start_new_run()
        manager._init_state.assert_called_once()
        MTurkManagerFile.mturk_utils.setup_sns_topic.assert_called_once_with(
            manager.opt['task'], manager.server_url, manager.task_group_id
        )
        self.assertEqual(manager.topic_arn, topic_arn)
        self.assertEqual(manager.task_state, manager.STATE_INIT_RUN)

        # connect to the server
        manager._setup_socket = mock.MagicMock()
        manager.ready_to_accept_workers()
        manager._setup_socket.assert_called_once()
        self.assertEqual(manager.task_state, MTurkManager.STATE_ACCEPTING_WORKERS)

        # 'launch' some hits
        manager.create_additional_hits = mock.MagicMock(return_value=mturk_page_url)
        hits_url = manager.create_hits()
        manager.create_additional_hits.assert_called_once()
        self.assertEqual(manager.task_state, MTurkManager.STATE_HITS_MADE)
        self.assertEqual(hits_url, mturk_page_url)

        # start a task
        manager.num_conversations = 10
        manager.expire_all_unassigned_hits = mock.MagicMock()
        manager._expire_onboarding_pool = mock.MagicMock()
        manager._expire_agent_pool = mock.MagicMock()

        # Run a task, ensure it closes when the max convs have been 'had'
        def run_task():
            manager.start_task(lambda worker: True, None, None)

        task_thread = threading.Thread(target=run_task, daemon=True)
        task_thread.start()

        self.assertTrue(task_thread.isAlive())
        manager.started_conversations = 10
        manager.completed_conversations = 10
        assert_equal_by(task_thread.isAlive, False, 0.6)
        manager.expire_all_unassigned_hits.assert_called_once()
        manager._expire_onboarding_pool.assert_called_once()
        manager._expire_agent_pool.assert_called_once()

        # shutdown
        manager.expire_all_unassigned_hits = mock.MagicMock()
        manager._expire_onboarding_pool = mock.MagicMock()
        manager._expire_agent_pool = mock.MagicMock()
        manager._wait_for_task_expirations = mock.MagicMock()
        MTurkManagerFile.mturk_utils.delete_sns_topic = mock.MagicMock()
        manager.shutdown()
        self.assertTrue(manager.is_shutdown)
        manager.expire_all_unassigned_hits.assert_called_once()
        manager._expire_onboarding_pool.assert_called_once()
        manager._expire_agent_pool.assert_called_once()
        manager._wait_for_task_expirations.assert_called_once()
        MTurkManagerFile.server_utils.delete_server.assert_called_once()
        MTurkManagerFile.mturk_utils.delete_sns_topic.assert_called_once_with(topic_arn)


class TestMTurkManagerConnectedFunctions(unittest.TestCase):
    """
    Semi-unit semi-integration tests on the more state-dependent MTurkManager
    functionality.
    """

    def setUp(self):
        self.fake_socket = MockSocket()
        time.sleep(0.1)
        argparser = ParlaiParser(False, False)
        argparser.add_parlai_data_path()
        argparser.add_mturk_args()
        self.opt = argparser.parse_args([])
        self.opt['task'] = 'unittest'
        self.opt['assignment_duration_in_seconds'] = 6
        self.mturk_agent_ids = ['mturk_agent_1', 'mturk_agent_2']
        self.mturk_manager = MTurkManager(
            opt=self.opt, mturk_agent_ids=self.mturk_agent_ids, is_test=True
        )
        self.mturk_manager._init_state()
        self.mturk_manager.port = self.fake_socket.port
        self.mturk_manager._onboard_new_agent = mock.MagicMock()
        self.mturk_manager._wait_for_task_expirations = mock.MagicMock()
        self.mturk_manager.task_group_id = 'TEST_GROUP_ID'
        self.mturk_manager.server_url = 'https://127.0.0.1'
        self.mturk_manager.task_state = self.mturk_manager.STATE_ACCEPTING_WORKERS
        self.mturk_manager._setup_socket()
        alive_packet = Packet(
            '',
            '',
            '',
            '',
            '',
            {
                'worker_id': TEST_WORKER_ID_1,
                'hit_id': TEST_HIT_ID_1,
                'assignment_id': TEST_ASSIGNMENT_ID_1,
                'conversation_id': None,
            },
            '',
        )
        self.mturk_manager._on_alive(alive_packet)
        alive_packet = Packet(
            '',
            '',
            '',
            '',
            '',
            {
                'worker_id': TEST_WORKER_ID_2,
                'hit_id': TEST_HIT_ID_2,
                'assignment_id': TEST_ASSIGNMENT_ID_2,
                'conversation_id': None,
            },
            '',
        )
        self.mturk_manager._on_alive(alive_packet)
        self.agent_1 = self.mturk_manager.worker_manager.get_agent_for_assignment(
            TEST_ASSIGNMENT_ID_1
        )
        self.agent_2 = self.mturk_manager.worker_manager.get_agent_for_assignment(
            TEST_ASSIGNMENT_ID_2
        )

    def tearDown(self):
        self.mturk_manager.shutdown()
        self.fake_socket.close()

    def test_socket_dead(self):
        """
        Test all states of socket dead calls.
        """
        manager = self.mturk_manager
        agent = self.agent_1
        worker_id = agent.worker_id
        assignment_id = agent.assignment_id
        manager.socket_manager.close_channel = mock.MagicMock()
        agent.reduce_state = mock.MagicMock()
        agent.set_status = mock.MagicMock(wraps=agent.set_status)
        manager._handle_agent_disconnect = mock.MagicMock(
            wraps=manager._handle_agent_disconnect
        )

        # Test status none
        agent.set_status(AssignState.STATUS_NONE)
        agent.set_status.reset_mock()
        manager._on_socket_dead(worker_id, assignment_id)
        self.assertEqual(agent.get_status(), AssignState.STATUS_DISCONNECT)
        agent.reduce_state.assert_called_once()
        manager.socket_manager.close_channel.assert_called_once_with(
            agent.get_connection_id()
        )
        manager._handle_agent_disconnect.assert_not_called()

        # Test status onboarding
        agent.set_status(AssignState.STATUS_ONBOARDING)
        agent.set_status.reset_mock()
        agent.reduce_state.reset_mock()
        manager.socket_manager.close_channel.reset_mock()
        self.assertFalse(agent.disconnected)
        manager._on_socket_dead(worker_id, assignment_id)
        self.assertEqual(agent.get_status(), AssignState.STATUS_DISCONNECT)
        agent.reduce_state.assert_called_once()
        manager.socket_manager.close_channel.assert_called_once_with(
            agent.get_connection_id()
        )
        self.assertTrue(agent.disconnected)
        manager._handle_agent_disconnect.assert_not_called()

        # test status waiting
        agent.disconnected = False
        agent.set_status(AssignState.STATUS_WAITING)
        agent.set_status.reset_mock()
        agent.reduce_state.reset_mock()
        manager.socket_manager.close_channel.reset_mock()
        manager._add_agent_to_pool(agent)
        manager._remove_from_agent_pool = mock.MagicMock()
        manager._on_socket_dead(worker_id, assignment_id)
        self.assertEqual(agent.get_status(), AssignState.STATUS_DISCONNECT)
        agent.reduce_state.assert_called_once()
        manager.socket_manager.close_channel.assert_called_once_with(
            agent.get_connection_id()
        )
        self.assertTrue(agent.disconnected)
        manager._handle_agent_disconnect.assert_not_called()
        manager._remove_from_agent_pool.assert_called_once_with(agent)

        # test status in task
        agent.disconnected = False
        agent.set_status(AssignState.STATUS_IN_TASK)
        agent.set_status.reset_mock()
        agent.reduce_state.reset_mock()
        manager.socket_manager.close_channel.reset_mock()
        manager._add_agent_to_pool(agent)
        manager._remove_from_agent_pool = mock.MagicMock()
        manager._on_socket_dead(worker_id, assignment_id)
        self.assertEqual(agent.get_status(), AssignState.STATUS_DISCONNECT)
        manager.socket_manager.close_channel.assert_called_once_with(
            agent.get_connection_id()
        )
        self.assertTrue(agent.disconnected)
        manager._handle_agent_disconnect.assert_called_once_with(
            worker_id, assignment_id
        )

        # test status done
        agent.disconnected = False
        agent.set_status(AssignState.STATUS_DONE)
        agent.set_status.reset_mock()
        agent.reduce_state.reset_mock()
        manager._handle_agent_disconnect.reset_mock()
        manager.socket_manager.close_channel.reset_mock()
        manager._add_agent_to_pool(agent)
        manager._remove_from_agent_pool = mock.MagicMock()
        manager._on_socket_dead(worker_id, assignment_id)
        self.assertNotEqual(agent.get_status(), AssignState.STATUS_DISCONNECT)
        agent.reduce_state.assert_not_called()
        manager.socket_manager.close_channel.assert_not_called()
        self.assertFalse(agent.disconnected)
        manager._handle_agent_disconnect.assert_not_called()

    def test_send_message_command(self):
        manager = self.mturk_manager
        agent = self.agent_1
        worker_id = self.agent_1.worker_id
        assignment_id = self.agent_1.assignment_id
        agent.set_last_command = mock.MagicMock()
        manager.socket_manager.queue_packet = mock.MagicMock()

        # Send a command
        data = {'text': data_model.COMMAND_SEND_MESSAGE}
        manager.send_command(worker_id, assignment_id, data)

        agent.set_last_command.assert_called_once_with(data)
        manager.socket_manager.queue_packet.assert_called_once()
        packet = manager.socket_manager.queue_packet.call_args[0][0]
        self.assertIsNotNone(packet.id)
        self.assertEqual(packet.type, Packet.TYPE_MESSAGE)
        self.assertEqual(packet.receiver_id, worker_id)
        self.assertEqual(packet.assignment_id, assignment_id)
        self.assertEqual(packet.data, data)
        self.assertEqual(packet.data['type'], data_model.MESSAGE_TYPE_COMMAND)

        # Send a message
        data = {'text': 'This is a test message'}
        agent.set_last_command.reset_mock()
        manager.socket_manager.queue_packet.reset_mock()
        message_id = manager.send_message(worker_id, assignment_id, data)
        agent.set_last_command.assert_not_called()
        manager.socket_manager.queue_packet.assert_called_once()
        packet = manager.socket_manager.queue_packet.call_args[0][0]
        self.assertIsNotNone(packet.id)
        self.assertEqual(packet.type, Packet.TYPE_MESSAGE)
        self.assertEqual(packet.receiver_id, worker_id)
        self.assertEqual(packet.assignment_id, assignment_id)
        self.assertNotEqual(packet.data, data)
        self.assertEqual(data['text'], packet.data['text'])
        self.assertEqual(packet.data['message_id'], message_id)
        self.assertEqual(packet.data['type'], data_model.MESSAGE_TYPE_MESSAGE)

    def test_free_workers(self):
        manager = self.mturk_manager
        manager.socket_manager.close_channel = mock.MagicMock()
        manager.free_workers([self.agent_1])
        manager.socket_manager.close_channel.assert_called_once_with(
            self.agent_1.get_connection_id()
        )

    def test_force_expire_hit(self):
        manager = self.mturk_manager
        agent = self.agent_1
        worker_id = agent.worker_id
        assignment_id = agent.assignment_id
        socket_manager = manager.socket_manager
        manager.send_command = mock.MagicMock()
        socket_manager.close_channel = mock.MagicMock()

        # Test expiring finished worker
        agent.set_status(AssignState.STATUS_DONE)
        manager.force_expire_hit(worker_id, assignment_id)
        manager.send_command.assert_not_called()
        socket_manager.close_channel.assert_not_called()
        self.assertEqual(agent.get_status(), AssignState.STATUS_DONE)

        # Test expiring not finished worker with default args
        agent.set_status(AssignState.STATUS_ONBOARDING)
        manager.force_expire_hit(worker_id, assignment_id)
        manager.send_command.assert_called_once()
        args = manager.send_command.call_args[0]
        used_worker_id, used_assignment_id, data = args[0], args[1], args[2]
        ack_func = manager.send_command.call_args[1]['ack_func']
        ack_func()
        self.assertEqual(worker_id, used_worker_id)
        self.assertEqual(assignment_id, used_assignment_id)
        self.assertEqual(data['text'], data_model.COMMAND_EXPIRE_HIT)
        self.assertEqual(agent.get_status(), AssignState.STATUS_EXPIRED)
        self.assertTrue(agent.hit_is_expired)
        self.assertIsNotNone(data['inactive_text'])
        socket_manager.close_channel.assert_called_once_with(agent.get_connection_id())

        # Test expiring not finished worker with custom arguments
        agent.set_status(AssignState.STATUS_ONBOARDING)
        agent.hit_is_expired = False
        manager.send_command = mock.MagicMock()
        socket_manager.close_channel = mock.MagicMock()
        special_disconnect_text = 'You were disconnected as part of a test'
        test_ack_function = mock.MagicMock()
        manager.force_expire_hit(
            worker_id,
            assignment_id,
            text=special_disconnect_text,
            ack_func=test_ack_function,
        )
        manager.send_command.assert_called_once()
        args = manager.send_command.call_args[0]
        used_worker_id, used_assignment_id, data = args[0], args[1], args[2]
        ack_func = manager.send_command.call_args[1]['ack_func']
        ack_func()
        self.assertEqual(worker_id, used_worker_id)
        self.assertEqual(assignment_id, used_assignment_id)
        self.assertEqual(data['text'], data_model.COMMAND_EXPIRE_HIT)
        self.assertEqual(agent.get_status(), AssignState.STATUS_EXPIRED)
        self.assertTrue(agent.hit_is_expired)
        self.assertEqual(data['inactive_text'], special_disconnect_text)
        socket_manager.close_channel.assert_called_once_with(agent.get_connection_id())
        test_ack_function.assert_called()

    def test_get_qualifications(self):
        manager = self.mturk_manager
        mturk_utils = MTurkManagerFile.mturk_utils
        mturk_utils.find_or_create_qualification = mock.MagicMock()

        # create a qualification list with nothing but a provided junk qual
        fake_qual = {
            'QualificationTypeId': 'fake_qual_id',
            'Comparator': 'DoesNotExist',
            'ActionsGuarded': 'DiscoverPreviewAndAccept',
        }
        qualifications = manager.get_qualification_list([fake_qual])
        self.assertListEqual(qualifications, [fake_qual])
        self.assertListEqual(manager.qualifications, [fake_qual])
        mturk_utils.find_or_create_qualification.assert_not_called()

        # Create a qualificaiton list using all the default types
        disconnect_qual_name = 'disconnect_qual_name'
        disconnect_qual_id = 'disconnect_qual_id'
        block_qual_name = 'block_qual_name'
        block_qual_id = 'block_qual_id'
        max_time_qual_name = 'max_time_qual_name'
        max_time_qual_id = 'max_time_qual_id'
        unique_qual_name = 'unique_qual_name'
        unique_qual_id = 'unique_qual_id'

        def return_qualifications(qual_name, _text, _sb):
            if qual_name == disconnect_qual_name:
                return disconnect_qual_id
            if qual_name == block_qual_name:
                return block_qual_id
            if qual_name == max_time_qual_name:
                return max_time_qual_id
            if qual_name == unique_qual_name:
                return unique_qual_id

        mturk_utils.find_or_create_qualification = return_qualifications
        manager.opt['disconnect_qualification'] = disconnect_qual_name
        manager.opt['block_qualification'] = block_qual_name
        manager.opt['max_time_qual'] = max_time_qual_name
        manager.opt['unique_qual_name'] = unique_qual_name
        manager.is_unique = True
        manager.has_time_limit = True
        manager.qualifications = None
        qualifications = manager.get_qualification_list()

        for qual in qualifications:
            self.assertEqual(qual['ActionsGuarded'], 'DiscoverPreviewAndAccept')
            self.assertEqual(qual['Comparator'], 'DoesNotExist')

        for qual_id in [
            disconnect_qual_id,
            block_qual_id,
            max_time_qual_id,
            unique_qual_id,
        ]:
            has_qual = False
            for qual in qualifications:
                if qual['QualificationTypeId'] == qual_id:
                    has_qual = True
                    break
            self.assertTrue(has_qual)

        self.assertListEqual(qualifications, manager.qualifications)

    def test_create_additional_hits(self):
        manager = self.mturk_manager
        manager.opt['hit_title'] = 'test_hit_title'
        manager.opt['hit_description'] = 'test_hit_description'
        manager.opt['hit_keywords'] = 'test_hit_keywords'
        manager.opt['reward'] = 0.1
        mturk_utils = MTurkManagerFile.mturk_utils
        fake_hit = 'fake_hit_type'
        mturk_utils.create_hit_type = mock.MagicMock(return_value=fake_hit)
        mturk_utils.subscribe_to_hits = mock.MagicMock()
        mturk_utils.create_hit_with_hit_type = mock.MagicMock(
            return_value=('page_url', 'hit_id', 'test_hit_response')
        )
        manager.server_url = 'test_url'
        manager.task_group_id = 'task_group_id'
        manager.topic_arn = 'topic_arn'
        mturk_chat_url = '{}/chat_index?task_group_id={}'.format(
            manager.server_url, manager.task_group_id
        )
        hit_url = manager.create_additional_hits(5)
        mturk_utils.create_hit_type.assert_called_once()
        mturk_utils.subscribe_to_hits.assert_called_with(
            fake_hit, manager.is_sandbox, manager.topic_arn
        )
        self.assertEqual(len(mturk_utils.create_hit_with_hit_type.call_args_list), 5)
        mturk_utils.create_hit_with_hit_type.assert_called_with(
            opt=manager.opt,
            page_url=mturk_chat_url,
            hit_type_id=fake_hit,
            num_assignments=1,
            is_sandbox=manager.is_sandbox,
        )
        self.assertEqual(len(manager.hit_id_list), 5)
        self.assertEqual(hit_url, 'page_url')

    def test_expire_all_hits(self):
        manager = self.mturk_manager
        worker_manager = manager.worker_manager
        completed_hit_id = 'completed'
        incomplete_1 = 'incomplete_1'
        incomplete_2 = 'incomplete_2'
        MTurkManagerFile.mturk_utils.expire_hit = mock.MagicMock()
        worker_manager.get_complete_hits = mock.MagicMock(
            return_value=[completed_hit_id]
        )
        manager.hit_id_list = [completed_hit_id, incomplete_1, incomplete_2]

        manager.expire_all_unassigned_hits()
        worker_manager.get_complete_hits.assert_called_once()
        expire_calls = MTurkManagerFile.mturk_utils.expire_hit.call_args_list
        self.assertEqual(len(expire_calls), 2)
        for hit in [incomplete_1, incomplete_2]:
            found = False
            for expire_call in expire_calls:
                if expire_call[0][1] == hit:
                    found = True
                    break
            self.assertTrue(found)

    def test_qualification_management(self):
        manager = self.mturk_manager
        test_qual_name = 'test_qual'
        other_qual_name = 'other_qual'
        test_qual_id = 'test_qual_id'
        worker_id = self.agent_1.worker_id
        mturk_utils = MTurkManagerFile.mturk_utils
        success_id = 'Success'

        def find_qualification(qual_name, _sandbox):
            if qual_name == test_qual_name:
                return test_qual_id
            return None

        mturk_utils.find_qualification = find_qualification
        mturk_utils.give_worker_qualification = mock.MagicMock()
        mturk_utils.remove_worker_qualification = mock.MagicMock()
        mturk_utils.find_or_create_qualification = mock.MagicMock(
            return_value=success_id
        )

        # Test give qualification
        manager.give_worker_qualification(worker_id, test_qual_name)
        mturk_utils.give_worker_qualification.assert_called_once_with(
            worker_id, test_qual_id, None, manager.is_sandbox
        )

        # Test revoke qualification
        manager.remove_worker_qualification(worker_id, test_qual_name)
        mturk_utils.remove_worker_qualification.assert_called_once_with(
            worker_id, test_qual_id, manager.is_sandbox, ''
        )

        # Test create qualification can exist
        result = manager.create_qualification(test_qual_name, '')
        self.assertEqual(result, success_id)

        # Test create qualification can't exist failure
        result = manager.create_qualification(test_qual_name, '', False)
        self.assertIsNone(result)

        # Test create qualification can't exist success
        result = manager.create_qualification(other_qual_name, '')
        self.assertEqual(result, success_id)

    def test_partner_disconnect(self):
        manager = self.mturk_manager
        manager.send_command = mock.MagicMock()
        self.agent_1.set_status(AssignState.STATUS_IN_TASK)
        manager._handle_partner_disconnect(self.agent_1)
        self.assertEqual(
            self.agent_1.get_status(), AssignState.STATUS_PARTNER_DISCONNECT
        )
        args = manager.send_command.call_args[0]
        worker_id, assignment_id, data = args[0], args[1], args[2]
        self.assertEqual(worker_id, self.agent_1.worker_id)
        self.assertEqual(assignment_id, self.agent_1.assignment_id)
        self.assertDictEqual(data, self.agent_1.get_inactive_command_data())

    @testing_utils.retry()
    def test_restore_state(self):
        manager = self.mturk_manager
        worker_manager = manager.worker_manager
        worker_manager.change_agent_conversation = mock.MagicMock()
        manager.send_command = mock.MagicMock()

        agent = self.agent_1
        agent.conversation_id = 'Test_conv_id'
        agent.id = 'test_agent_id'
        agent.request_message = mock.MagicMock()
        agent.message_request_time = time.time()
        test_message = {
            'text': 'this_is_a_message',
            'message_id': 'test_id',
            'type': data_model.MESSAGE_TYPE_MESSAGE,
        }
        agent.append_message(test_message)
        manager._restore_agent_state(agent.worker_id, agent.assignment_id)
        self.assertFalse(agent.alived)
        manager.send_command.assert_not_called()
        worker_manager.change_agent_conversation.assert_called_once_with(
            agent=agent, conversation_id=agent.conversation_id, new_agent_id=agent.id
        )
        agent.alived = True
        assert_equal_by(lambda: len(agent.request_message.call_args_list), 1, 0.6)
        manager.send_command.assert_called_once()
        args = manager.send_command.call_args[0]
        worker_id, assignment_id, data = args[0], args[1], args[2]
        self.assertEqual(worker_id, agent.worker_id)
        self.assertEqual(assignment_id, agent.assignment_id)
        self.assertListEqual(data['messages'], agent.get_messages())
        self.assertEqual(data['text'], data_model.COMMAND_RESTORE_STATE)

    def test_expire_onboarding(self):
        manager = self.mturk_manager
        manager.force_expire_hit = mock.MagicMock()
        self.agent_2.set_status(AssignState.STATUS_ONBOARDING)
        manager._expire_onboarding_pool()

        manager.force_expire_hit.assert_called_once_with(
            self.agent_2.worker_id, self.agent_2.assignment_id
        )


if __name__ == '__main__':
    unittest.main(buffer=True)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Base tokenizer/tokens classes and utilities.
"""

import copy


class Tokens(object):
    """
    A class to represent a list of tokenized text.
    """

    TEXT = 0
    TEXT_WS = 1
    SPAN = 2
    POS = 3
    LEMMA = 4
    NER = 5

    def __init__(self, data, annotators, opts=None):
        self.data = data
        self.annotators = annotators
        self.opts = opts or {}

    def __len__(self):
        """
        The number of tokens.
        """
        return len(self.data)

    def slice(self, i=None, j=None):
        """
        Return a view of the list of tokens from [i, j).
        """
        new_tokens = copy.copy(self)
        new_tokens.data = self.data[i:j]
        return new_tokens

    def untokenize(self):
        """
        Returns the original text (with whitespace reinserted).
        """
        return ''.join([t[self.TEXT_WS] for t in self.data]).strip()

    def words(self, uncased=False):
        """
        Returns a list of the text of each token.

        Args:
            uncased: lower cases text
        """
        if uncased:
            return [t[self.TEXT].lower() for t in self.data]
        else:
            return [t[self.TEXT] for t in self.data]

    def offsets(self):
        """
        Returns a list of [start, end) character offsets of each token.
        """
        return [t[self.SPAN] for t in self.data]

    def pos(self):
        """
        Returns a list of part-of-speech tags of each token.

        Returns None if this annotation was not included.
        """
        if 'pos' not in self.annotators:
            return None
        return [t[self.POS] for t in self.data]

    def lemmas(self):
        """
        Returns a list of the lemmatized text of each token.

        Returns None if this annotation was not included.
        """
        if 'lemma' not in self.annotators:
            return None
        return [t[self.LEMMA] for t in self.data]

    def entities(self):
        """
        Returns a list of named-entity-recognition tags of each token.

        Returns None if this annotation was not included.
        """
        if 'ner' not in self.annotators:
            return None
        return [t[self.NER] for t in self.data]

    def ngrams(self, n=1, uncased=False, filter_fn=None, as_strings=True):
        """
        Returns a list of all ngrams from length 1 to n.

        Args:
            n: upper limit of ngram length
            uncased: lower cases text
            filter_fn: user function that takes in an ngram list and returns
              True or False to keep or not keep the ngram
            as_string: return the ngram as a string vs list
        """

        def _skip(gram):
            if not filter_fn:
                return False
            return filter_fn(gram)

        words = self.words(uncased)
        ngrams = [
            (s, e + 1)
            for s in range(len(words))
            for e in range(s, min(s + n, len(words)))
            if not _skip(words[s : e + 1])
        ]

        # Concatenate into strings
        if as_strings:
            ngrams = ['{}'.format(' '.join(words[s:e])) for (s, e) in ngrams]

        return ngrams

    def entity_groups(self):
        """
        Group consecutive entity tokens with the same NER tag.
        """
        entities = self.entities()
        if not entities:
            return None
        non_ent = self.opts.get('non_ent', 'O')
        groups = []
        idx = 0
        while idx < len(entities):
            ner_tag = entities[idx]
            # Check for entity tag
            if ner_tag != non_ent:
                # Chomp the sequence
                start = idx
                while idx < len(entities) and entities[idx] == ner_tag:
                    idx += 1
                groups.append((self.slice(start, idx).untokenize(), ner_tag))
            else:
                idx += 1
        return groups


class Tokenizer(object):
    """
    Base tokenizer class.

    Tokenizers implement tokenize, which should return a Tokens class.
    """

    def tokenize(self, text):
        raise NotImplementedError

    def shutdown(self):
        pass

    def __del__(self):
        self.shutdown()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
A self-feeding chatbot (a chatbot wrapped with an NRC and QGen)

A self-feeding chatbot consists of three components:
- dialog agent: this is a typical model for handlin a conversastion(e.g. Transformer)
- feedback classifier: this classifies incoming user responses
- question generator: when the feedback classifier predicts a sufficiently negative
    feedback score, this module asks for an feedback from the from the user.
"""

import random

import torch
import torch.nn as nn
import torch.nn.functional as F

from parlai.core.torch_agent import TorchAgent, Output
from parlai.utils.misc import round_sigfigs, warn_once
from parlai.utils.torch import padded_tensor

from parlai.agents.transformer.transformer import TransformerRankerAgent

from .feedback_classifier.feedback_classifier import FeedbackClassifierRegex
from .modules import SelfFeedingModel

EPS = 1e-9

# Status options
NORMAL = 1
FEEDBACK_REQUESTED = 2
NEWTOPIC_REQUESTED = 3
RATING_REQUESTED = 4
RATING_ACCEPTED = 5

# Hardcoded responses
RAT_REQUEST = (
    "Just checking: I'm not sure how positive/negative the response you "
    "just gave was. Could you help me out? (Write positive, negative, or "
    "neutral)"
)
CONTINUE = "And in response to what you were saying before"
FEEDBACK_REQUEST = (
    "Oops! I think I messed up. Whether I messed up or not, what could I "
    "have said (in response to <response>)?"
)
THANKS = "Thanks! I'll try to remember that."
NEWTOPIC = "Can you pick a new topic for us to talk about now?"


class SelfFeedingAgent(TransformerRankerAgent):
    @classmethod
    def add_cmdline_args(cls, argparser):
        """
        Add command-line arguments specifically for this agent.
        """
        super().add_cmdline_args(argparser)
        SelfFeedingModel.add_cmdline_args(argparser)

        agent = argparser.add_argument_group('Self-feeding Agent')
        agent.add_argument(
            '--request-rating',
            type='bool',
            default=False,
            help="If True, ocassionally request ratings",
        )
        agent.add_argument(
            '--rating-frequency',
            type=float,
            default=0.01,
            help="The fraction of the time that a rating will be "
            "randomly requested, regardless of classifier confidence",
        )
        agent.add_argument(
            '--rating-gap',
            type=float,
            default=0.05,
            help="A positivity rating must be within this amount of "
            "the rating-threshold to trigger the confidence-based "
            "rating request",
        )

        agent.add_argument(
            '--request-feedback',
            type='bool',
            default=False,
            help="If True, recognize mistakes and request feedback",
        )
        agent.add_argument(
            '--rating-threshold',
            type=float,
            default=0.5,
            help="Treat feedback below this threshold as negative",
        )
        agent.add_argument(
            '--display-sat-estimate',
            type='bool',
            default=False,
            help="If True, print estimated satisfaction after each response",
        )
        agent.add_argument(
            '--target-class',
            type=int,
            default=0,
            help="The label (in [0,1]) to treat as the target rarer "
            "class when calculating tp/fp/tn/fn",
        )

        agent.add_argument(
            '--freeze-base',
            type='bool',
            default=False,
            help="If True, freeze all but the satisfaction linear layer",
        )
        agent.add_argument(
            '--partial-load',
            type='bool',
            default=False,
            help="If True, allow a model to be partially loaded (i.e., "
            "a model need not be a perfect match to be loaded",
        )

        agent.add_argument(
            '-subtasks',
            '--subtasks',
            type=str,
            help="A comma-separated list of active subtasks",
        )

        agent.add_argument(
            '--dia-weight',
            type=float,
            default=1.0,
            help="The dialog task loss is multiplied by this",
        )
        agent.add_argument(
            '--fee-weight',
            type=float,
            default=1.0,
            help="The feedback task loss is multiplied by this",
        )
        agent.add_argument(
            '--sat-weight',
            type=float,
            default=1.0,
            help="The satisfaction task loss is multiplied by this",
        )

        agent.add_argument(
            '--prev-response-negatives',
            type='bool',
            default=False,
            help="If True, during training add the previous agent response as a "
            "negative candidate for the current response (as a way to "
            "organically teach the model to not repeat itself). This option is "
            "ignored when candidate set is 'fixed' or 'vocab', as these already "
            "include all candidates",
        )
        agent.add_argument(
            '--prev-response-filter',
            type='bool',
            default=False,
            help="If True and --interactive=True, do not allow the model to output its "
            "previous response. This is a hackier solution than using "
            "--prev-response-negatives, but MUCH faster/simpler",
        )
        agent.add_argument(
            '--interactive',
            type='bool',
            default=False,
            help="Mark as true if you are in a setting where you are only doing "
            "evaluation, and always with the same fixed candidate set.",
        )

        agent.add_argument(
            '--add-double-person-tokens',
            type='bool',
            default=False,
            help="For backwards compatibility with old models only. For new "
            "models, this should always be False.",
        )

        agent.set_defaults(person_tokens=False)  # add these in a special way

        variants = argparser.add_argument_group('Self-feeding Variants')
        variants.add_argument(
            '-rgx',
            '--regex',
            type='bool',
            default=False,
            help="If True, classify satisfaction using regexes instead " "of model",
        )
        variants.add_argument(
            '-up',
            '--uncertainty-predictor',
            type='bool',
            default=False,
            help="If True, classify satisfaction using uncertainty of "
            "dialog models predictions instead of classifer"
            "model",
        )
        variants.add_argument(
            '-ut',
            '--uncertainty-threshold',
            type=float,
            default=0.5,
            help="If model confidence is smaller than this number and"
            " --uncertainty-predictor=True, classify as bot mistake",
        )
        variants.add_argument(
            '-us',
            '--uncertainty-style',
            type=str,
            default='gap',
            choices=['gap', 'mag'],
            help="Whether the uncertainty threshold measures the "
            "magnitude of the top confidence, or the gap between the "
            "two most confident answers",
        )
        return agent

    def __init__(self, opt, shared=None):
        if opt['interactive']:
            print("[ Setting interactive mode defaults... ]")
            opt['prev_response_filter'] = True

        # Set subtasks first so that opt['subtasks'] is set before build_model()
        self.set_subtasks(opt)
        self.multitask = len(self.subtasks) > 1
        if not self.multitask:
            self.subtask = self.subtasks[0]

        if opt['prev_response_negatives']:
            if opt['candidates'] in ['fixed', 'vocab']:
                msg = (
                    "[ Option --prev-response-negatives=True is incompatible with "
                    "--candidates=['fixed','vocab']. Overriding it to False. ]"
                )
                warn_once(msg)
                self.opt['prev_response_negatives'] = False
            self.prev_responses = None
        if opt['prev_response_filter']:
            if not opt['interactive']:
                msg = (
                    "[ Option --prev-response-filter=True can only be used when "
                    "--interactive=True. Overriding it to False. ]"
                )
                warn_once(msg)
                self.opt['prev_response_filter'] = False
            self.prev_response = None

        super().__init__(opt, shared)

        self.status = NORMAL
        if self.opt['interactive']:
            assert 'dialog' in self.subtasks
            assert 'satisfaction' in self.subtasks
        else:
            assert not self.opt['request_feedback']
            assert not self.opt['request_rating']

        self.task_weight = {
            'dialog': opt['dia_weight'],
            'feedback': opt['fee_weight'],
            'satisfaction': opt['sat_weight'],
        }

        # dialog/feedback tasks use self.rank_loss from TorchRankerAgent
        # Don't do BCEWithLogitsLoss since we need the probs from the sigmoid anyway
        self.satisfaction_criterion = nn.BCELoss(reduce=True, size_average=False)

        # Set rating classifier
        if opt['regex']:
            self.rating_classifier = FeedbackClassifierRegex()

        random.seed()
        self.reset()

    @classmethod
    def upgrade_opt(cls, opt_on_disk):
        """
        Upgrade opts from older model files.
        """
        super(SelfFeedingAgent, cls).upgrade_opt(opt_on_disk)

        # 2019-06-25: previous versions of the model did not add a CLS token
        # to the beginning of text_vec.
        if 'add_double_person_tokens' not in opt_on_disk:
            warn_once('Old model: overriding `add_double_person_tokens` to True.')
            opt_on_disk['add_double_person_tokens'] = True

        return opt_on_disk

    # NOTE: This is the only method of TransformerAgent being overwritten
    def build_model(self):
        self.model = SelfFeedingModel(self.opt, self.dict)
        if self.opt['embedding_type'] != 'random':
            for embeddings in [
                getattr(self.model, 'dia_embeddings', None),
                getattr(self.model, 'fee_embeddings', None),
                getattr(self.model, 'sat_embeddings', None),
            ]:
                if embeddings is not None:
                    self._copy_embeddings(embeddings.weight, self.opt['embedding_type'])
        return self.model

    def act(self):
        if self.opt['request_feedback'] or self.opt['request_rating']:
            # rating request ->
            #   0/+1 -> normal
            #   -1 -> feedback request
            # feedback request -> topic request -> normal
            positivity = -1  # actual positivity values are in [0, 1]

            # Preprocessing (for special states that transition to NORMAL)
            if self.status in [NEWTOPIC_REQUESTED, RATING_ACCEPTED]:
                self.status = NORMAL

            if self.status == NORMAL:
                positivity = self.predict_satisfaction(self.observation)
                if self.opt["display_sat_estimate"]:
                    print(f"[ satisfaction estimate ]: {positivity}")
                if self.do_request_feedback(positivity):
                    action = self.make_action(self.make_feedback_request())
                    self.status = FEEDBACK_REQUESTED
                elif self.do_request_rating(positivity):
                    action = self.make_action(self.make_rating_request())
                    self.requested_rating = True
                    self.status = RATING_REQUESTED

            elif self.status == RATING_REQUESTED:
                rating = self.extract_rating()
                if rating == -1:
                    action = self.make_action(
                        self.make_feedback_request(), reward=rating
                    )
                    self.status = FEEDBACK_REQUESTED

                else:
                    action, reply = self.make_rating_response(rating)
                    self.status = RATING_ACCEPTED
                    # Store just the non-template part of the response in the history
                    self.history.update_history({"text": reply})

            elif self.status == FEEDBACK_REQUESTED:
                action = self.make_action(THANKS + ' ' + NEWTOPIC)
                self.status = NEWTOPIC_REQUESTED
                self.history.reset()

            else:
                raise Exception(f"Unrecognized status: {self.status}")

        if self.status == NORMAL:
            action = super().act()
            self.history.update_history(action)

        return action

    def observe(self, observation):
        """
        Add to history, concatenate history-size utterances, and add person tokens.
        """

        # If their response is a response to a rating request, no work required
        if self.status == RATING_REQUESTED:
            self.last_rating = observation['text']

        return super().observe(observation)

    def batchify(self, observations):
        batch = super().batchify(observations)
        if self.multitask and batch.text_vec is not None:
            # Dialog is the default task; e.g., in interactive mode
            subtasks = [o.get('subtask', 'dialog') for o in observations]
            # Catch error here for debugging
            assert len(set(subtasks)) == 1
            self.subtask = subtasks[0]
        # print(f"[ context ]: {observations[0]['text']}")
        return batch

    def _set_text_vec(self, obs, history, truncate):
        """
        Set the 'text_vec' field in the observation.
        """
        if 'text' not in obs:
            return obs

        if 'text_vec' not in obs:
            obs['text_vec'] = self.history.parse(obs["text"])

        # check truncation
        if 'text_vec' in obs:
            if self.opt.get('add_double_person_tokens'):
                # backwards compatibility madness, handle the fact that __p1__ was
                # accidentally inserted into the training data of some old models on
                # disk
                obs['text_vec'].insert(0, self.dict['__p1__'])
            obs.force_set(
                'text_vec',
                torch.LongTensor(self._check_truncate(obs['text_vec'], truncate, True)),
            )

        return obs

    def train_step(self, batch):
        """
        Train on a single batch of examples.
        """
        if batch.text_vec is None:
            return

        self.model.train()
        self.optimizer.zero_grad()

        batchsize = batch.text_vec.size(0)
        self.metrics['examples'] += batchsize

        if self.subtask == 'dialog':
            loss, preds, _ = self.dialog_step(batch)
        elif self.subtask == 'feedback':
            loss, preds, _ = self.feedback_step(batch)
        elif self.subtask == 'satisfaction':
            loss, preds = self.satisfaction_step(batch)
            preds = [str(p) for p in preds]

        # Weight loss by task-weight
        loss *= self.task_weight[self.subtask]

        loss.backward()
        self.update_params()

        return Output(preds)

    def eval_step(self, batch):
        if batch.text_vec is None:
            return

        self.model.eval()

        batchsize = batch.text_vec.size(0)
        self.metrics['examples'] += batchsize

        if self.subtask == 'dialog':
            _, preds, cand_ranked = self.dialog_step(batch)
            if self.opt['interactive']:
                if self.opt['prev_response_filter']:
                    preds = self.check_prev_response(preds, cand_ranked)
                return Output(preds)
            else:
                return Output(preds, cand_ranked)

        elif self.subtask == 'feedback':
            _, preds, cand_ranked = self.feedback_step(batch)
            return Output(preds, cand_ranked)

        elif self.subtask == 'satisfaction':
            if self.opt['uncertainty_predictor']:
                # Use uncertainty of dialog model to classify bot's previous utterance
                preds = self.predict_satisfaction_by_uncertainty(batch)
            else:
                # Use satisfaction of user response to classify bot's previous response
                _, preds = self.satisfaction_step(batch)
            preds = [str(p) for p in preds]
            return Output(preds)

    def dialog_step(self, batch):
        batchsize = batch.text_vec.size(0)
        if self.model.training:
            cands, cand_vecs, label_inds = self._build_candidates(
                batch, source=self.opt['candidates'], mode='train'
            )
        else:
            cands, cand_vecs, label_inds = self._build_candidates(
                batch, source=self.opt['eval_candidates'], mode='eval'
            )
            if self.encode_candidate_vecs and self.eval_candidates in [
                'fixed',
                'vocab',
            ]:
                # if we cached candidate encodings for a fixed list of candidates,
                # pass those into the score_candidates function
                if self.eval_candidates == 'fixed':
                    cand_vecs = self.fixed_candidate_encs
                elif self.eval_candidates == 'vocab':
                    cand_vecs = self.vocab_candidate_encs

        scores = self.model.score_dialog(batch.text_vec, cand_vecs)
        _, ranks = scores.sort(1, descending=True)

        if self.model.training:
            # Get predictions but not full rankings for the sake of speed
            cand_ranked = None
            if cand_vecs.dim() == 2:
                preds = [cands[ordering[0]] for ordering in ranks]
            elif cand_vecs.dim() == 3:
                preds = [cands[i][ordering[0]] for i, ordering in enumerate(ranks)]
        else:
            # Return full rankings to calculate hits@ metrics
            cand_ranked = []
            for i, ordering in enumerate(ranks):
                if cand_vecs.dim() == 2:
                    cand_list = cands
                elif cand_vecs.dim() == 3:
                    cand_list = cands[i]
                cand_ranked.append([cand_list[rank] for rank in ordering])
            preds = [cand_ranked[i][0] for i in range(batchsize)]

        if label_inds is None:
            loss = None
        else:
            loss = self.criterion(scores, label_inds).mean()
            self.update_dia_metrics(loss, ranks, label_inds, batchsize)
        return loss, preds, cand_ranked

    def feedback_step(self, batch):
        batchsize = batch.text_vec.size(0)

        warn_once("WARNING: feedback candidates are hardcoded to batch")
        if self.model.training:
            cands, cand_vecs, label_inds = self._build_candidates(
                batch, source='batch', mode='train'
            )
        else:
            cands, cand_vecs, label_inds = self._build_candidates(
                batch, source='batch', mode='eval'
            )

        scores = self.model.score_feedback(batch.text_vec, cand_vecs)
        _, ranks = scores.sort(1, descending=True)

        if self.model.training:
            # Get predictions but not full rankings for the sake of speed
            cand_ranked = None
            preds = [cands[ordering[0]] for ordering in ranks]
        else:
            # Return full rankings to calculate hits@ metrics
            cand_ranked = []
            for ordering in ranks:
                cand_ranked.append([cands[rank] for rank in ordering])
            preds = [cand_ranked[i][0] for i in range(batchsize)]

        if label_inds is None:
            loss = None
        else:
            loss = self.criterion(scores, label_inds).mean()
            self.update_fee_metrics(loss, ranks, label_inds, batchsize)
        return loss, preds, cand_ranked

    def satisfaction_step(self, batch):
        batchsize = batch.text_vec.size(0)
        # probs is a [batchsize] torch.FloatTensor with values in [0, 1]
        if self.opt['regex']:
            contexts = [self.dict.vec2txt(vec) for vec in batch.text_vec]
            # HACK: --use-cuda flag
            probs = self.rating_classifier.predict_proba(contexts).cuda()
        else:
            probs = self.model.score_satisfaction(batch.text_vec)
        # preds is a [batchsize] torch.LongTensor with values in {0, 1}
        preds = (probs > self.opt['rating_threshold']).long()

        if batch.labels is None:  # TODO: not sure if it's None, list of Nones, etc.
            loss = None
        else:
            # labels will be a [batchsize] torch.LongTensor with values in {0, 1}
            labels = torch.LongTensor([int(l) == 1 for l in batch.labels]).cuda()

            loss = self.satisfaction_criterion(probs, labels.float()).mean()
            self.update_sat_metrics(loss, preds, labels, batchsize)

        return loss, preds

    def update_dia_metrics(self, loss, ranks, label_inds, batchsize):
        self.metrics['dia_exs'] += batchsize
        self.metrics['dia_loss'] += loss.item()
        if label_inds is not None:
            for b in range(batchsize):
                rank = (ranks[b] == label_inds[b]).nonzero().item()
                self.metrics['dia_rank'] += 1 + rank
                self.metrics['dia_correct'] += rank == 0

    def update_fee_metrics(self, loss, ranks, label_inds, batchsize):
        self.metrics['fee_exs'] += batchsize
        self.metrics['fee_loss'] += loss.item()
        if label_inds is not None:
            for b in range(batchsize):
                rank = (ranks[b] == label_inds[b]).nonzero().item()
                self.metrics['fee_rank'] += 1 + rank
                self.metrics['fee_correct'] += rank == 0

    def update_sat_metrics(self, loss, preds, labels, batchsize):
        # tp/fp/tn/fn are w/r/t the rarer class (negative responses)
        self.metrics['sat_exs'] += batchsize
        self.metrics['sat_loss'] += loss.item()
        a = self.opt['target_class']
        b = not self.opt['target_class']
        assert a in [0, 1]
        assert b in [0, 1]
        self.metrics['sat_tp'] += ((preds == labels) * (labels == a)).sum().item()
        self.metrics['sat_fp'] += ((preds != labels) * (labels == b)).sum().item()
        self.metrics['sat_tn'] += ((preds == labels) * (labels == b)).sum().item()
        self.metrics['sat_fn'] += ((preds != labels) * (labels == a)).sum().item()

    def reset(self):
        super().reset()
        self.requested_rating = False

    def reset_metrics(self):
        """
        Reset metrics.
        """
        super().reset_metrics()
        self.metrics['examples'] = 0
        if 'dialog' in self.subtasks:
            self.metrics['dia_exs'] = 0
            self.metrics['dia_loss'] = 0.0
            self.metrics['dia_rank'] = 0
            self.metrics['dia_correct'] = 0
        if 'feedback' in self.subtasks:
            self.metrics['fee_exs'] = 0
            self.metrics['fee_loss'] = 0.0
            self.metrics['fee_rank'] = 0
            self.metrics['fee_correct'] = 0.0
        if 'satisfaction' in self.subtasks:
            self.metrics['sat_exs'] = 0
            self.metrics['sat_loss'] = 0
            self.metrics['sat_tp'] = 0
            self.metrics['sat_fp'] = 0
            self.metrics['sat_tn'] = 0
            self.metrics['sat_fn'] = 0

    def report(self):
        """
        Report metrics from model's perspective.
        """
        m = TorchAgent.report(self)  # Skip TorchRankerAgent; totally redundant
        examples = self.metrics['examples']
        if examples > 0:
            m['examples'] = examples
            if 'dialog' in self.subtasks and self.metrics['dia_exs'] > 0:
                m['dia_loss'] = self.metrics['dia_loss'] / self.metrics['dia_exs']
                m['dia_rank'] = self.metrics['dia_rank'] / self.metrics['dia_exs']
                m['dia_acc'] = self.metrics['dia_correct'] / self.metrics['dia_exs']
                m['dia_exs'] = self.metrics['dia_exs']
            if 'feedback' in self.subtasks and self.metrics['fee_exs'] > 0:
                m['fee_loss'] = self.metrics['fee_loss'] / self.metrics['fee_exs']
                m['fee_rank'] = self.metrics['fee_rank'] / self.metrics['fee_exs']
                m['fee_acc'] = self.metrics['fee_correct'] / self.metrics['fee_exs']
                m['fee_exs'] = self.metrics['fee_exs']
                m['fee_exs'] = self.metrics['fee_exs']
            if 'satisfaction' in self.subtasks and self.metrics['sat_exs'] > 0:
                tp = self.metrics['sat_tp']
                tn = self.metrics['sat_tn']
                fp = self.metrics['sat_fp']
                fn = self.metrics['sat_fn']
                assert tp + tn + fp + fn == self.metrics['sat_exs']
                m['sat_loss'] = self.metrics['sat_loss'] / self.metrics['sat_exs']
                m['sat_pr'] = tp / (tp + fp + EPS)
                m['sat_re'] = tp / (tp + fn + EPS)
                pr = m['sat_pr']
                re = m['sat_re']
                m['sat_f1'] = (2 * pr * re) / (pr + re) if (pr and re) else 0.0
                m['sat_acc'] = (tp + tn) / self.metrics['sat_exs']
                m['sat_exs'] = self.metrics['sat_exs']
        for k, v in m.items():
            # clean up: rounds to sigfigs and converts tensors to floats
            if isinstance(v, float):
                m[k] = round_sigfigs(v, 4)
            else:
                m[k] = v
        return m

    def encode_candidates(self, cands):
        """
        Encodes a tensor of vectorized candidates.

        :param cands: a [bs, seq_len] or [bs, num_cands, seq_len](?) of vectorized
            candidates
        """
        return self.model.encode_dia_y(cands)

    def do_request_feedback(self, positivity):
        """
        Decide whether to request an feedback this turn.
        """
        # If --request-feedback=False, then don't request an feedback
        if not self.opt['request_feedback'] or len(self.history.history_strings) == 1:
            return False
        else:
            return positivity < self.opt['rating_threshold']

    def do_request_rating(self, positivity):
        """
        Decide whether to request a rating this turn.
        """
        # If --request-rating=False, then don't request a rating
        if not self.opt['request_rating']:
            return False
        # If no previous bot utterance is on record, can't ask about it
        elif len(self.history.history_strings) < 2:
            return False
        # Only request a rating a maximum of once per episode
        elif self.requested_rating:
            return False
        # Some percentage of the time, we ask for a rating randomly (to aid exploration)
        elif random.random() < self.opt['rating_frequency']:
            return True
        # Lastly, ask for a rating based on confidence (proximity to threshold)
        else:
            gap = abs(positivity - self.opt['rating_threshold'])
            return gap < self.opt['rating_gap']

    def extract_rating(self):
        """
        Convert user response to rating request from text to an integer rating.
        """
        # TODO: Make this more robust!
        if self.last_rating == 'positive':
            return 1
        elif self.last_rating == 'negative':
            return -1
        else:
            return 0

    def predict_satisfaction(self, observation):
        if self.opt['regex']:
            prob = self.rating_classifier.predict_proba(observation['text_vec'])
        else:
            prob = self.model.score_satisfaction(observation['text_vec'].reshape(1, -1))
        return prob.item()

    def predict_satisfaction_by_uncertainty(self, batch):
        # Use the dialog model's confidence to predict the rating on previous response
        # HACK: this test is run using a model that was trained on dialog but is now
        # being evaluated on satisfaction. We use a satisfaction dataset so that we have
        # access to the satisfaction labels. Therefore, we do a sloppy hack here to pull
        # out what would have been the candidates for the dialog task. We use histsz=4
        # and ignore the last response (the user's feedback) and use the penultimate
        # utterances as the candidates. The (up to) two utterances before that are
        # context.

        # pull out dialog candidates from text_vecs since this is a satisfaction task
        assert self.opt['history_size'] > 2
        text_vecs = []
        cand_vecs = []
        for vec in batch.text_vec:
            last_p1 = (vec == self.dict.txt2vec('__p1__')[0]).nonzero()[-1].item()
            last_p2 = (vec == self.dict.txt2vec('__p2__')[0]).nonzero()[-1].item()
            text_vecs.append(vec[:last_p2])
            cand_vecs.append(vec[last_p2 + 1 : last_p1])
        text_padded, _ = padded_tensor(text_vecs)
        cand_padded, _ = padded_tensor(cand_vecs)
        scores = self.model.score_dialog(text_padded, cand_padded)
        confidences = F.softmax(scores, dim=1).cpu().detach().numpy()

        preds = []
        for example in confidences:
            ranked_confidences = sorted(list(example), reverse=True)
            if self.opt['uncertainty_style'] == 'mag':
                # If the most confident choice isn't confident enough, predict that
                # the response the bot gives will be bad (pred=0)
                mag = ranked_confidences[0]
                preds.append(mag > self.opt['uncertainty_threshold'])
            elif self.opt['uncertainty_style'] == 'gap':
                # If the gap between the first and second most confident choices isn't
                # large enough, predict that the response the bot gives will be bad
                # (pred=0)
                gap = ranked_confidences[0] - ranked_confidences[1]
                preds.append(gap > self.opt['uncertainty_threshold'])

        loss = torch.tensor(0)
        preds = torch.LongTensor(preds)
        labels = torch.LongTensor([int(l) == 1 for l in batch.labels])
        batchsize = len(labels)
        self.update_sat_metrics(loss, preds, labels, batchsize)
        return preds

    def check_prev_response(self, preds, cand_ranked):
        # Compare current prediction to previous (replacing if necessary)
        if self.prev_response and (preds[0] == self.prev_response):
            preds[0] = cand_ranked[0][1]
        # Save current prediction for next turn
        self.prev_response = preds[0]
        return preds

    def make_action(self, text, reward=0):
        action = {'id': self.id, 'text': text}
        if reward:
            action['reward'] = reward
        return action

    def make_feedback_request(self):
        return 'Oops! I think I messed up. ' 'What could I have said instead?'

    def make_rating_request(self):
        return RAT_REQUEST

    def make_rating_response(self, rating):
        action = super().act()
        reply = str(action['text'])
        action['reward'] = rating
        last_message = self.history.history_strings[-1]
        if rating == 0:
            action['text'] = f'Okay, thanks! {CONTINUE} ("{last_message}"): {reply}'
        elif rating == 1:
            action['text'] = f'Great, thanks! {CONTINUE} ("{last_message}"): {reply}'
        return action, reply

    def set_subtasks(self, opt):
        # Find assigned subtasks
        if opt.get('subtasks', None):
            # First check if it was passed explicitly as a flag or otherwise in opt
            if isinstance(opt['subtasks'], list):
                subtasks = opt['subtasks']
            else:
                subtasks = opt['subtasks'].split(',')
        else:
            # Otherwise, try to infer from the task name.
            subtasks = [task.split(':')[1] for task in opt['task'].split(',')]

        # Expand any abbreviations
        if subtasks[0] == 'diafee':
            subtasks = ['dialog', 'feedback']
        elif subtasks[0] == 'diasat':
            subtasks = ['dialog', 'satisfaction']
        elif subtasks[0] == 'all':
            subtasks = ['dialog', 'feedback', 'satisfaction']

        self.subtasks = subtasks
        opt['subtasks'] = subtasks  # Add to opt so that model module can see the list

    def load(self, path):
        """
        Return opt and model states.

        Overriding TorchAgent.load() to enable partial loading
        """
        states = torch.load(path, map_location=lambda cpu, _: cpu)
        if 'model' in states:
            try:
                self.model.load_state_dict(states['model'])
            except RuntimeError as e:
                if self.opt['partial_load']:
                    print(
                        "WARNING: could not load entire --init-model; "
                        "loading partially instead."
                    )
                    pretrained_state = states['model']
                    current_state = self.model.state_dict()
                    # 1. filter out unnecessary keys
                    pretrained_dict = {
                        k: v for k, v in pretrained_state.items() if k in current_state
                    }
                    # 2. overwrite entries in the existing state dict
                    current_state.update(pretrained_dict)
                    # 3. load the new state dict
                    self.model.load_state_dict(current_state)
                else:
                    raise Exception(
                        f"The designated model could not be loaded. "
                        f"Consider using --partial-load=True.\n {e}"
                    )

        if 'optimizer' in states and hasattr(self, 'optimizer'):
            self.optimizer.load_state_dict(states['optimizer'])
        return states

    def _add_prev_responses(self, batch, cands, cand_vecs, label_inds, source):
        assert source not in ['fixed', 'vocab']
        self._extract_prev_responses(batch)

        # Add prev_responses as negatives
        prev_cands = self.prev_responses
        prev_cand_vecs = [torch.Tensor(self.dict.txt2vec(cand)) for cand in prev_cands]
        if source == 'batch':
            # Option 1: Change from one set of shared candidates to separate per example
            # cands = [cands + [prev_cand] for prev_cand in prev_cands]
            # list_of_lists_of_cand_vecs = [[vec for vec in cand_vecs] + [prev_cand_vec]
            #                            for prev_cand_vec in prev_cand_vecs]
            # cand_vecs = padded_3d(list_of_lists_of_cand_vecs, use_cuda=self.use_cuda,
            #                       dtype=cand_vecs[0].dtype)

            # Option 2: Just add all prev responses for the whole batch (doubles bs)
            cands += prev_cands
            cand_vecs, _ = padded_tensor(
                [vec for vec in cand_vecs] + prev_cand_vecs, use_cuda=self.use_cuda
            )
        elif source == 'inline':
            raise NotImplementedError

        return cands, cand_vecs

    def _extract_prev_responses(self, batch):
        # Extract prev_responses for self-feeding formatted examples
        warn_once("WARNING: This code is specific to self-feeding formatted examples")

        # TODO: Pull out p1/p2 once elsewhere, not every time
        p1 = self.dict.txt2vec('__p1__')[0]
        p2 = self.dict.txt2vec('__p2__')[0]
        self.prev_responses = []
        # Do naively for now with a for loop
        for text_vec in batch.text_vec:
            p1s = (text_vec == p1).nonzero()
            p2s = (text_vec == p2).nonzero()
            if len(p1s) and len(p2s):
                response_vec = text_vec[p2s[-1] + 1 : p1s[-1]]
            else:
                response_vec = [self.NULL_IDX]  # TODO: pull in actual N
            response = self.dict.vec2txt(response_vec)
            self.prev_responses.append(response)

    def _build_candidates(self, batch, source, mode):
        cands, cand_vecs, label_inds = super()._build_candidates(batch, source, mode)
        if self.opt['prev_response_negatives'] and mode == 'train':
            cands, cand_vecs = self._add_prev_responses(
                batch, cands, cand_vecs, label_inds, source
            )
        return cands, cand_vecs, label_inds
